from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Generic

from cognite.client.data_classes import (
    Asset,
    AssetList,
    AssetWrite,
    AssetWriteList,
    DataSetWriteList,
    FileMetadata,
    FileMetadataList,
    FileMetadataWrite,
    FileMetadataWriteList,
    Label,
    LabelDefinition,
)
from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
)
from cognite.client.data_classes.labels import LabelDefinitionWriteList
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.cruds import AssetCRUD, DataSetsCRUD, FileMetadataCRUD, LabelCRUD, ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.utils.aggregators import AssetAggregator, AssetCentricAggregator, FileAggregator
from cognite_toolkit._cdf_tk.utils.cdf import metadata_key_counts
from cognite_toolkit._cdf_tk.utils.file import find_files_with_suffix_and_prefix
from cognite_toolkit._cdf_tk.utils.fileio import SchemaColumn
from cognite_toolkit._cdf_tk.utils.useful_types import T_ID, JsonVal, T_WritableCogniteResourceList

from ._base import StorageIOConfig, TableStorageIO
from ._selectors import AssetCentricFileSelector, AssetCentricSelector, AssetSubtreeSelector, DataSetSelector


class BaseAssetCentricIO(
    Generic[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList],
    TableStorageIO[int, AssetCentricSelector, T_CogniteResourceList, T_WritableCogniteResourceList],
    ABC,
):
    chunk_size = 1000

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._loader = self._get_loader()
        self._aggregator = self._get_aggregator()
        self._downloaded_data_sets_by_selector: dict[AssetCentricSelector, set[int]] = defaultdict(set)
        self._downloaded_labels_by_selector: dict[AssetCentricSelector, set[str]] = defaultdict(set)

    def as_id(self, item: dict[str, JsonVal] | object) -> int:
        if isinstance(item, dict) and isinstance(item.get("id"), int):
            # MyPy checked above.
            return item["id"]  # type: ignore[return-value]
        raise TypeError(f"Cannot extract ID from item of type {type(item).__name__!r}")

    @abstractmethod
    def _get_loader(
        self,
    ) -> ResourceCRUD[
        T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList
    ]:
        raise NotImplementedError()

    @abstractmethod
    def _get_aggregator(self) -> AssetCentricAggregator:
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, ids: Sequence[int]) -> T_WritableCogniteResourceList:
        raise NotImplementedError()

    def count(self, selector: AssetCentricSelector) -> int | None:
        if isinstance(selector, DataSetSelector):
            return self._aggregator.count(data_set_external_id=selector.data_set_external_id)
        elif isinstance(selector, AssetSubtreeSelector):
            return self._aggregator.count(hierarchy=selector.hierarchy)
        return None

    def data_to_json_chunk(self, data_chunk: T_WritableCogniteResourceList) -> list[dict[str, JsonVal]]:
        return [self._loader.dump_resource(item) for item in data_chunk]

    def configurations(self, selector: AssetCentricSelector) -> Iterable[StorageIOConfig]:
        data_set_ids = self._downloaded_data_sets_by_selector[selector]
        if data_set_ids:
            data_set_external_ids = self.client.lookup.data_sets.external_id(list(data_set_ids))
            yield from self._configurations(data_set_external_ids, DataSetsCRUD.create_loader(self.client))

        yield from self._configurations(
            list(self._downloaded_labels_by_selector[selector]), LabelCRUD.create_loader(self.client)
        )

    def _collect_dependencies(self, resources: AssetList | FileMetadataList, selector: AssetCentricSelector) -> None:
        for resource in resources:
            if resource.data_set_id:
                self._downloaded_data_sets_by_selector[selector].add(resource.data_set_id)
            for label in resource.labels or []:
                if isinstance(label, str):
                    self._downloaded_labels_by_selector[selector].add(label)
                elif isinstance(label, Label | LabelDefinition) and label.external_id:
                    self._downloaded_labels_by_selector[selector].add(label.external_id)
                elif isinstance(label, dict) and "externalId" in label:
                    self._downloaded_labels_by_selector[selector].add(label["externalId"])

    @classmethod
    def _configurations(
        cls,
        ids: list[str],
        loader: DataSetsCRUD | LabelCRUD,
    ) -> Iterable[StorageIOConfig]:
        if not ids:
            return
        items = loader.retrieve(list(ids))
        yield StorageIOConfig(
            kind=loader.kind,
            folder_name=loader.folder_name,
            # We know that the items will be labels for LabelLoader and data sets for DataSetsLoader
            value=[loader.dump_resource(item) for item in items],  # type: ignore[arg-type]
        )

    def load_selector(self, datafile: Path) -> AssetCentricSelector:
        return AssetCentricFileSelector(datafile=datafile)

    def ensure_configurations(self, selector: AssetCentricSelector, console: Console | None = None) -> None:
        """Ensures that all data sets and labels referenced by the asset selection exist in CDF."""
        if not isinstance(selector, AssetCentricFileSelector):
            return None
        datafile = selector.datafile
        filepaths = find_files_with_suffix_and_prefix(
            datafile.parent.parent / DataSetsCRUD.folder_name, datafile.name, suffix=f".{DataSetsCRUD.kind}.yaml"
        )
        self._create_if_not_exists(filepaths, DataSetsCRUD.create_loader(self.client), console)

        filepaths = find_files_with_suffix_and_prefix(
            datafile.parent.parent / LabelCRUD.folder_name, datafile.name, suffix=f".{LabelCRUD.kind}.yaml"
        )
        self._create_if_not_exists(filepaths, LabelCRUD.create_loader(self.client), console)
        return None

    @classmethod
    def _create_if_not_exists(
        cls,
        filepaths: list[Path],
        loader: DataSetsCRUD | LabelCRUD,
        console: Console | None = None,
    ) -> None:
        items: LabelDefinitionWriteList | DataSetWriteList = loader.list_write_cls([])
        for filepath in filepaths:
            if not filepath.exists():
                continue
            for loaded in loader.load_resource_file(filepath):
                items.append(loader.load_resource(loaded))
        # MyPy fails to understand that existing, existing_ids and missing will be consistent for given loader.
        existing = loader.retrieve(loader.get_ids(items))
        existing_ids = set(loader.get_ids(existing))
        if missing := [item for item in items if loader.get_id(item) not in existing_ids]:  # type: ignore[arg-type]
            loader.create(loader.list_write_cls(missing))  # type: ignore[arg-type]
            if console:
                console.print(
                    f"Created {loader.kind} for {len(missing)} items: {', '.join(str(item) for item in loader.get_ids(missing))}"  # type: ignore[arg-type]
                )


class AssetIO(BaseAssetCentricIO[str, AssetWrite, Asset, AssetWriteList, AssetList]):
    folder_name = "classic"
    kind = "Assets"
    display_name = "Assets"
    supported_download_formats = frozenset({".parquet", ".csv", ".ndjson"})
    supported_compressions = frozenset({".gz"})
    supported_read_formats = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})

    def as_id(self, item: dict[str, JsonVal] | object) -> int:
        if isinstance(item, Asset | AssetWrite) and item.id is not None:  # type: ignore[union-attr]
            return item.id  # type: ignore[union-attr]
        return super().as_id(item)

    def _get_loader(self) -> AssetCRUD:
        return AssetCRUD.create_loader(self.client)

    def _get_aggregator(self) -> AssetCentricAggregator:
        return AssetAggregator(self.client)

    def get_schema(self, selector: AssetCentricSelector) -> list[SchemaColumn]:
        data_set_ids: list[int] = []
        if isinstance(selector, DataSetSelector):
            data_set_ids.append(self.client.lookup.data_sets.id(selector.data_set_external_id))
        hierarchy: list[int] = []
        if isinstance(selector, AssetSubtreeSelector):
            hierarchy.append(self.client.lookup.assets.id(selector.hierarchy))

        if hierarchy or data_set_ids:
            metadata_keys = metadata_key_counts(
                self.client, "assets", data_sets=data_set_ids or None, hierarchies=hierarchy or None
            )
        else:
            metadata_keys = []
        metadata_schema: list[SchemaColumn] = []
        if metadata_keys:
            metadata_schema.extend(
                [SchemaColumn(name=f"metadata.{key}", type="string", is_array=False) for key, _ in metadata_keys]
            )
        asset_schema = [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="name", type="string"),
            SchemaColumn(name="parentExternalId", type="string"),
            SchemaColumn(name="description", type="string"),
            SchemaColumn(name="dataSetExternalId", type="string"),
            SchemaColumn(name="source", type="string"),
            SchemaColumn(name="labels", type="string", is_array=True),
            SchemaColumn(name="geoLocation", type="json"),
        ]
        return asset_schema + metadata_schema

    def download_iterable(self, selector: AssetCentricSelector, limit: int | None = None) -> Iterable[AssetList]:
        asset_subtree_external_ids: list[str] | None = None
        data_set_external_ids: list[str] | None = None
        if isinstance(selector, DataSetSelector):
            data_set_external_ids = [selector.data_set_external_id]
        elif isinstance(selector, AssetSubtreeSelector):
            asset_subtree_external_ids = [selector.hierarchy]
        else:
            # This selector is for uploads, not for downloading from CDF.
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for AssetIO.")
        for asset_list in self.client.assets(
            chunk_size=self.chunk_size,
            limit=limit,
            asset_subtree_external_ids=asset_subtree_external_ids,
            data_set_external_ids=data_set_external_ids,
        ):
            self._collect_dependencies(asset_list, selector)
            yield asset_list

    def upload_items(self, data_chunk: AssetWriteList, selector: AssetCentricSelector) -> None:
        if not data_chunk:
            return
        self.client.assets.upsert(data_chunk, mode="patch")

    def json_chunk_to_data(self, data_chunk: list[dict[str, JsonVal]]) -> AssetWriteList:
        return AssetWriteList([self._loader.load_resource(item) for item in data_chunk])

    def retrieve(self, ids: Sequence[int]) -> AssetList:
        return self.client.assets.retrieve_multiple(ids)


class FileMetadataIO(BaseAssetCentricIO[str, FileMetadataWrite, FileMetadata, FileMetadataWriteList, FileMetadataList]):
    folder_name = FileMetadataCRUD.folder_name
    kind = "FileMetadata"
    display_name = "file metadata"
    supported_download_formats = frozenset({".parquet", ".csv", ".ndjson"})
    supported_compressions = frozenset({".gz"})
    supported_read_formats = frozenset({".parquet", ".csv", ".ndjson"})

    def as_id(self, item: dict[str, JsonVal] | object) -> int:
        if isinstance(item, FileMetadata | FileMetadataWrite) and item.id is not None:  # type: ignore[union-attr]
            return item.id  # type: ignore[union-attr]
        return super().as_id(item)

    def _get_loader(self) -> FileMetadataCRUD:
        return FileMetadataCRUD.create_loader(self.client)

    def _get_aggregator(self) -> AssetCentricAggregator:
        return FileAggregator(self.client)

    def get_schema(self, selector: AssetCentricSelector) -> list[SchemaColumn]:
        data_set_ids: list[int] = []
        if isinstance(selector, DataSetSelector):
            data_set_ids.append(self.client.lookup.data_sets.id(selector.data_set_external_id))
        if isinstance(selector, AssetSubtreeSelector):
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for FileIO.")

        if data_set_ids:
            metadata_keys = metadata_key_counts(self.client, "files", data_sets=data_set_ids or None, hierarchies=None)
        else:
            metadata_keys = []
        metadata_schema: list[SchemaColumn] = []
        if metadata_keys:
            metadata_schema.extend(
                [SchemaColumn(name=f"metadata.{key}", type="string", is_array=False) for key, _ in metadata_keys]
            )
        file_schema = [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="name", type="string"),
            SchemaColumn(name="directory", type="string"),
            SchemaColumn(name="mimeType", type="string"),
            SchemaColumn(name="dataSetExternalId", type="string"),
            SchemaColumn(name="assetExternalIds", type="string", is_array=True),
            SchemaColumn(name="source", type="string"),
            SchemaColumn(name="sourceCreatedTime", type="integer"),
            SchemaColumn(name="sourceModifiedTime", type="integer"),
            SchemaColumn(name="securityCategories", type="string", is_array=True),
            SchemaColumn(name="labels", type="string", is_array=True),
            SchemaColumn(name="geoLocation", type="json"),
        ]
        return file_schema + metadata_schema

    def download_iterable(self, selector: AssetCentricSelector, limit: int | None = None) -> Iterable[FileMetadataList]:
        asset_subtree_external_ids: list[str] | None = None
        data_set_external_ids: list[str] | None = None
        if isinstance(selector, DataSetSelector):
            data_set_external_ids = [selector.data_set_external_id]
        elif isinstance(selector, AssetSubtreeSelector):
            asset_subtree_external_ids = [selector.hierarchy]
        else:
            # This selector is for uploads, not for downloading from CDF.
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for FileMetadataIO.")
        for file_list in self.client.files(
            chunk_size=self.chunk_size,
            limit=limit,
            asset_subtree_external_ids=asset_subtree_external_ids,
            data_set_external_ids=data_set_external_ids,
        ):
            self._collect_dependencies(file_list, selector)
            yield file_list

    def retrieve(self, ids: Sequence[int]) -> FileMetadataList:
        return self.client.files.retrieve_multiple(ids)

    def upload_items(self, data_chunk: FileMetadataWriteList, selector: AssetCentricSelector) -> None:
        if not data_chunk:
            return
        self._loader.create(data_chunk)

    def json_chunk_to_data(self, data_chunk: list[dict[str, JsonVal]]) -> FileMetadataWriteList:
        return FileMetadataWriteList([self._loader.load_resource(item) for item in data_chunk])



import json

import pytest
import requests
import responses
from cognite.client.data_classes.data_modeling import EdgeApply, NodeApply, ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstanceApplyList
from cognite_toolkit._cdf_tk.storageio import InstanceIO, InstanceViewSelector
from cognite_toolkit._cdf_tk.utils.http_client import FailedItem, HTTPClient, SuccessItem


class TestInstanceIO:
    def test_download_instance_ids(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = toolkit_config.create_api_url("/models/instances/list")
        selector = InstanceViewSelector(
            ViewId("mySpace", "myView", "v42"), instance_type="node", instance_spaces=("my_insta_space",)
        )
        N = 2500
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                toolkit_config.create_api_url("/models/instances/aggregate"),
                status=200,
                json={
                    "items": [
                        {
                            "instanceType": "node",
                            "aggregates": [{"aggregate": "count", "property": "externalId", "value": N}],
                        }
                    ]
                },
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={
                    "items": [
                        {
                            "externalId": f"instance_{i}",
                            "space": "my_space",
                            "instanceType": "node",
                            "createdTime": 0,
                            "lastUpdatedTime": 0,
                            "version": 1,
                        }
                        for i in range(N)
                    ]
                },
            )
            io = InstanceIO(client)
            ids = list(io.download_ids(selector))
            count = io.count(selector)
        assert len(list(ids)) == N // io.chunk_size + (1 if N % io.chunk_size > 0 else 0)
        total_ids = sum(len(chunk) for chunk in ids)
        assert total_ids == N
        assert count == N

    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_upload_force(self, toolkit_config: ToolkitClientConfig, rsps: responses.RequestsMock) -> None:
        config = toolkit_config
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        instance_count = 12
        with HTTPClient(config) as http_client:
            instances = InstanceApplyList(
                [
                    NodeApply(
                        space="my_space",
                        external_id=f"instance_{i}",
                    )
                    if i % 2 == 0
                    else EdgeApply(
                        space="my_space",
                        external_id=f"instance_{i}",
                        type=("schema_space", "edge_type"),
                        start_node=("my_space", f"instance_{i - 1}"),
                        end_node=("my_space", f"instance_{i + 1}"),
                    )
                    for i in range(instance_count)
                ]
            )

            def hate_edges(request: requests.PreparedRequest) -> tuple[int, dict[str, str], str]:
                # status, headers, body
                if "edge" in request.body:
                    return 400, {}, '{"error":{"code":"InvalidArgument","message":"I do not like edges!"}}'
                else:
                    items = json.loads(request.body).get("items", [])
                    return (
                        200,
                        {},
                        json.dumps(
                            {
                                "items": [
                                    {
                                        "instanceType": "node",
                                        "space": item["space"],
                                        "externalId": item["externalId"],
                                        "wasModified": True,
                                        "createdTime": 0,
                                        "lastUpdatedTime": 0,
                                    }
                                    for item in items
                                ]
                            }
                        ),
                    )

            url = toolkit_config.create_api_url("/models/instances")

            rsps.add_callback(
                responses.POST,
                url,
                callback=hate_edges,
                content_type="application/json",
            )
            io = InstanceIO(client)
            results = io.upload_items_force(instances, http_client)

            assert len(results) == instance_count
            failed_items = [res for res in results if isinstance(res, FailedItem)]
            assert len(failed_items) == instance_count // 2
            success_items = [res for res in results if isinstance(res, SuccessItem)]
            assert len(success_items) == instance_count // 2



import gzip
import random
import time
from collections.abc import MutableMapping

import requests
from cognite.client import ClientConfig, CogniteClient, global_config
from cognite.client._api.functions import FunctionsAPI
from cognite.client._http_client import HTTPClient, HTTPClientConfig, get_global_requests_session
from cognite.client.data_classes import Function, FunctionWrite
from cognite.client.utils import _json
from cognite.client.utils._auxiliary import get_user_agent
from requests.structures import CaseInsensitiveDict
from rich.console import Console

from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils._auxiliary import get_current_toolkit_version
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


class ExtendedFunctionsAPI(FunctionsAPI):
    def __init__(
        self,
        config: ClientConfig,
        api_version: str | None,
        cognite_client: CogniteClient,
    ) -> None:
        """
        Extended Functions API to include custom headers and payload preparation.
        """
        super().__init__(config, api_version, cognite_client)
        session = get_global_requests_session()
        # The HTTPClient in the parent class always retries 429 responses, but we want to handle that manually.
        self._http_client_no_retry = HTTPClient(
            config=HTTPClientConfig(
                status_codes_to_retry=set(),
                backoff_factor=0.5,
                max_backoff_seconds=global_config.max_retry_backoff,
                max_retries_total=global_config.max_retries,
                max_retries_read=0,
                max_retries_connect=global_config.max_retries_connect,
                max_retries_status=0,
            ),
            session=session,
            refresh_auth_header=self._refresh_auth_header,
        )

    def create_with_429_retry(self, function: FunctionWrite, console: Console | None = None) -> Function:
        """Create a function with manual retry handling for 429 Too Many Requests responses.

        This method is a workaround for scenarios where the function creation API is temporarily unavailable
        due to a full queue on the server side.

        Args:
            function (FunctionWrite): The function to create.
            console (Console | None): The rich console to use for printing warnings.

        Returns:
            Function: The created function object.
        """
        headers = self._create_headers()
        payload = self._prepare_payload({"items": [function.dump(camel_case=True)]})
        _, full_url = self._resolve_url("POST", self._RESOURCE_PATH)
        retry_count = 0
        while True:
            response = self._http_client_no_retry.request(
                method="POST",
                url=full_url,
                headers=headers,
                data=payload,
                timeout=self._config.timeout,
                allow_redirects=False,
            )

            match response.status_code:
                case 200 | 201 | 202 | 204:
                    return Function._load(response.json()["items"][0], cognite_client=self._cognite_client)
                case 401:
                    self._raise_no_project_access_error(response)
                case 429 if retry_count < global_config.max_retries:
                    try:
                        retry_after = float(response.headers.get("Retry-After", 60))
                    except ValueError:
                        retry_after = 60.0 + random.uniform(-5, 5)
                    HighSeverityWarning(f"Rate limit exceeded. Retrying after {retry_after} seconds.").print_warning(
                        console=console
                    )
                    time.sleep(retry_after)
                    retry_count += 1
                    continue
                case _:
                    self._raise_api_error(response, {"items": function.dump(camel_case=True)})

    def _create_headers(self) -> MutableMapping[str, str]:
        headers: MutableMapping[str, str] = CaseInsensitiveDict()
        headers.update(requests.utils.default_headers())
        auth_name, auth_value = self._config.credentials.authorization_header()
        headers[auth_name] = auth_value
        headers["content-type"] = "application/json"
        headers["accept"] = "application/json"
        headers["x-cdp-sdk"] = f"CogniteToolkit:{get_current_toolkit_version()}"
        headers["x-cdp-app"] = self._config.client_name
        headers["cdf-version"] = self._config.api_subversion
        if "User-Agent" in headers:
            headers["User-Agent"] += f" {get_user_agent()}"
        else:
            headers["User-Agent"] = get_user_agent()
        if not global_config.disable_gzip:
            headers["Content-Encoding"] = "gzip"
        return headers

    @staticmethod
    def _prepare_payload(body: JsonVal) -> str | bytes:
        """
        Prepare the payload for the HTTP request.
        This method should be overridden in subclasses to customize the payload format.
        """
        data: str | bytes
        try:
            data = _json.dumps(body, allow_nan=False)
        except ValueError as e:
            # A lot of work to give a more human friendly error message when nans and infs are present:
            msg = "Out of range float values are not JSON compliant"
            if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
                raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                    e.__traceback__
                ) from None
            raise

        if not global_config.disable_gzip:
            data = gzip.compress(data.encode())
        return data



import threading
import time
from collections.abc import Iterator
from unittest.mock import MagicMock, patch

import pytest
import requests
import responses
from cognite.client import global_config
from cognite.client.credentials import OAuthClientCredentials

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils.auth import CLIENT_NAME
from cognite_toolkit._cdf_tk.utils.batch_processor import (
    BatchResult,
    HTTPBatchProcessor,
    HTTPIterableProcessor,
    SuccessItem,
)


@pytest.fixture
def toolkit_config() -> ToolkitClientConfig:
    global_config.disable_pypi_version_check = True
    credentials = MagicMock(spec=OAuthClientCredentials)
    credentials.client_id = "toolkit-client-id"
    credentials.client_secret = "toolkit-client-secret"
    credentials.token_url = "https://toolkit.auth.com/oauth/token"
    credentials.scopes = ["ttps://pytest-field.cognitedata.com/.default"]
    credentials.authorization_header.return_value = "Auth", "Bearer dummy"
    return ToolkitClientConfig(
        client_name=CLIENT_NAME,
        project="pytest-project",
        credentials=credentials,
        is_strict_validation=False,
    )


@pytest.fixture
def processor(toolkit_config: ToolkitClientConfig) -> Iterator[HTTPIterableProcessor]:
    with HTTPIterableProcessor(
        endpoint_url="https://test.com/api",
        config=toolkit_config,
        as_id=lambda item: item["id"],
        max_workers=4,
        batch_size=100,
    ) as processor:
        yield processor


class TestHTTPIterableProcessor:
    def test_happy_path(self, toolkit_config: ToolkitClientConfig) -> None:
        url = "http://example.com/api"
        processor: HTTPIterableProcessor[str]
        with (
            HTTPIterableProcessor[str](
                url,
                toolkit_config,
                lambda item: item["externalId"],
            ) as processor,
            responses.RequestsMock() as rsps,
        ):
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={
                    "items": [
                        {"externalId": "item1", "server": "property1"},
                        {"externalId": "item2", "server": "property2"},
                    ]
                },
            )
            result = processor.process(
                (
                    {
                        "externalId": "item1",
                    },
                    {"externalId": "item2"},
                )
            )

            assert len(rsps.calls) == 1
            assert result.total_successful == 2
            assert result.total_processed == 2
            assert result.success_rate == 1.0
            assert [item.response for item in result.successful_items if item.response] == [
                {"externalId": "item1", "server": "property1"},
                {"externalId": "item2", "server": "property2"},
            ]

    def test_concurrent_processing(self, processor: HTTPIterableProcessor) -> None:
        """Test that items are processed concurrently using multiple workers"""
        with responses.RequestsMock() as rsps:
            # Add a callback to simulate delay and track request times
            request_times = []

            def request_callback(request):
                request_times.append(time.time())
                time.sleep(0.1)
                return 200, {}, '{"items": []}'

            rsps.add_callback(responses.POST, processor.endpoint_url, callback=request_callback)

            items = [{"id": i, "name": f"Item {i}"} for i in range(400)]

            start_time = time.time()
            result = processor.process(items, total_items=len(items))
            end_time = time.time()

            assert len(request_times) == processor.max_workers
            # With 4 workers, we should have at least some parallel requests
            # Check that requests were made in parallel by verifying multiple requests
            # were initiated before the first one completed
            first_request_complete = request_times[0] + 0.1  # 0.1s delay per request
            parallel_requests = sum(1 for t in request_times if t < first_request_complete)
            assert parallel_requests > 1

            # Verify the total time is less than if it were sequential
            assert end_time - start_time < processor.max_workers * 0.1
            assert result.total_processed == 400

    @pytest.mark.usefixtures("disable_gzip")
    def test_split_batch_concurrency(self, processor: HTTPIterableProcessor) -> None:
        """Test batch splitting under concurrent load"""
        processed_batch_counts = []

        def request_callback(request):
            # Extract the batch size from the request payload
            batch_count = request.body.count('"id":')
            processed_batch_counts.append(batch_count)

            # First batch (size 100) gets an error, others succeed
            if batch_count == 100:
                return 502, {}, '{"error": {"message": "Server Error"}}'
            else:
                return 200, {}, request.body

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                responses.POST,
                processor.endpoint_url,
                callback=request_callback,
            )

            items = [{"id": i} for i in range(100)]
            result = processor.process(items, total_items=len(items))

            assert processed_batch_counts == [100, 50, 50], "Batch should be split into 100, 50, 50"
            assert result.total_successful == 100

    @pytest.mark.usefixtures("disable_gzip")
    def test_network_errors(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test prevention of queue deadlocks under error conditions"""
        url = "https://test.com/api"

        def connection_error_callback(request):
            if '"id": 1' in request.body:
                # Fist worker
                raise requests.exceptions.ReadTimeout("Read timeout error")
            else:
                raise requests.exceptions.ConnectionError("Connection error")

        with (
            HTTPIterableProcessor(
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                max_workers=2,
                batch_size=25,
                max_retries=2,  # Small max_retries to prevent test from running too long
            ) as processor,
            responses.RequestsMock() as rsps,
        ):
            rsps.add_callback(
                responses.POST,
                url,
                callback=connection_error_callback,
            )

            with patch("time.sleep"):  # Skip actual waiting
                items = [{"id": i} for i in range(50)]
                result = processor.process(items, total_items=len(items))

                # All items should eventually fail after retries
                assert result.total_failed == 50
                # Each item should have been attempted max_retries
                # With 50 items and batch size 25, that's 2 batches x max_retries
                assert len(rsps.calls) == 2 * processor.max_workers

    def test_worker_shutdown(self, toolkit_config):
        """Test that workers properly terminate when signaled"""
        url = "https://test.com/api"
        # For this test, we need to monitor the worker threads directly
        # We'll use a mock for the worker method to detect shutdown signals
        shutdown_signal_count = 0

        def mock_worker(work_queue, results_queue):
            nonlocal shutdown_signal_count
            while True:
                work_item = work_queue.get()
                if work_item is None:
                    shutdown_signal_count += 1
                    work_queue.task_done()
                    break
                results_queue.put(BatchResult(successful_items=[SuccessItem("id", 200) for _ in work_item.items]))
                work_queue.task_done()

        with (
            HTTPIterableProcessor(
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                max_workers=4,
                batch_size=25,
            ) as processor,
            patch.object(processor, "_worker", side_effect=mock_worker),
        ):
            items = [{"id": i} for i in range(50)]
            processor.process(items, total_items=len(items))

            # Verify each worker received a shutdown signal
            assert shutdown_signal_count == processor.max_workers

    def test_no_access(self, processor: HTTPIterableProcessor) -> None:
        """Test handling of 401 Unauthorized responses"""
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                processor.endpoint_url,
                status=401,
                json={"error": {"message": "Unauthorized access"}},
            )

            items = [{"id": i} for i in range(10)]
            result = processor.process(items, total_items=len(items))

            assert result.total_failed == 10
            assert result.total_successful == 0
            assert 401 in result.error_summary
            assert result.error_summary[401] == 10
            # Verify there was only one request attempt (no retries for 401)
            assert len(rsps.calls) == 1

    def test_rate_limit_handling(self, processor: HTTPIterableProcessor) -> None:
        """Test handling of 429 rate limit responses with eventual success"""
        items = [{"id": i} for i in range(10)]
        responses_sequence = [
            # First attempt gets rate limited
            {"status": 429, "json": {"error": {"message": "Too many requests"}}},
            # Second attempt succeeds
            {"status": 200, "json": {"items": items}},
        ]

        with responses.RequestsMock() as rsps:
            for resp in responses_sequence:
                rsps.add(
                    responses.POST,
                    processor.endpoint_url,
                    status=resp["status"],
                    json=resp["json"],
                )

            with patch("time.sleep"):  # Skip actual waiting
                result = processor.process(items, total_items=len(items))

                assert result.total_successful == 10
                assert result.total_failed == 0
                assert len(rsps.calls) == 2  # First fails, second succeeds

                # Verify rate limit was handled
                assert rsps.calls[0].response.status_code == 429
                assert rsps.calls[1].response.status_code == 200

    @pytest.mark.usefixtures("disable_gzip")
    def test_permanent_failure_single_item(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test handling of permanent failure for a single item"""
        url = "https://test.com/api"
        # Track which items were sent in requests
        request_items = []

        def request_callback(request):
            payload = request.body.decode() if isinstance(request.body, bytes) else request.body
            # Extract the item ID from the request
            if '"id": 3' in payload:  # Item with ID 3 will fail
                return 400, {}, '{"error": {"message": "Invalid format for item"}}'
            else:
                request_items.append(payload)
                # Return payload for successful items
                return 200, {}, payload

        with (
            HTTPIterableProcessor(
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                max_workers=2,
                batch_size=1,
            ) as processor,
            responses.RequestsMock() as rsps,
        ):
            rsps.add_callback(
                responses.POST,
                url,
                callback=request_callback,
            )

            items = [{"id": i} for i in range(5)]
            result = processor.process(items, total_items=len(items))

            # Verify one item failed and four succeeded
            assert result.total_successful == 4
            assert result.total_failed == 1
            assert 400 in result.error_summary
            assert result.error_summary[400] == 1

            # Verify the failed item has ID 3
            failed_id = result.failed_items[0].item
            assert failed_id == 3

            # Verify each item was processed individually
            assert len(rsps.calls) == 5

    def test_invalid_as_id_function(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test that an error is raised if the as_id function is invalid"""

        def invalid_as_id(item) -> str:
            raise ValueError("Invalid ID function")

        with (
            responses.RequestsMock() as rsps,
            HTTPIterableProcessor[str](
                endpoint_url="https://test.com/api",
                config=toolkit_config,
                as_id=invalid_as_id,
                max_workers=2,
                batch_size=10,
            ) as processor,
        ):
            rsps.add(
                responses.POST,
                processor.endpoint_url,
                status=200,
                json={"items": []},
            )
            result = processor.process([{"id": 1}, {"id": 2}, {"id": 3}])

            assert len(result.unknown_items) == 3

    def test_empty_batch(self, processor: HTTPIterableProcessor) -> None:
        """Test that processing an empty batch returns an empty result"""
        result = processor.process([])

        assert result.total_successful == 0
        assert result.total_failed == 0
        assert result.total_processed == 0

    def test_raise_in_iteration(self, processor: HTTPIterableProcessor) -> None:
        """Test that an exception raised during iteration is handled correctly"""

        def items_generator():
            yield {"id": 1}
            yield {"id": 2}
            raise ValueError("Raising an error during iteration")

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                processor.endpoint_url,
                status=200,
                json={"items": []},
            )
            result = processor.process(items_generator())

        assert str(result.producer_error) == "Raising an error during iteration"


class TestHTTPBatchProcessor:
    def test_happy_path(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test that TestHTTPBatchProcessor processes items correctly"""
        url = "https://test.com/api"
        items = [{"id": i} for i in range(5)]

        batches: list[BatchResult[str]] = []

        def processor(batch: BatchResult[str]) -> None:
            batches.append(batch)

        with (
            responses.RequestsMock() as rsps,
            HTTPBatchProcessor[str](
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                result_processor=processor,
                max_workers=2,
            ) as processor,
        ):
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 1}, {"id": 2}, {"id": 3}]},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 4}, {"id": 5}]},
            )
            processor.add_items(items[:3])
            processor.add_items(items[3:])

        assert len(batches) == 2

    def test_iterable_processor_empty_input(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test that TestHTTPBatchProcessor does not call result_processor for empty input"""
        url = "https://test.com/api"
        batches: list[BatchResult[str]] = []

        def processor(batch: BatchResult[str]) -> None:
            batches.append(batch)

        with (
            responses.RequestsMock() as _,
            HTTPBatchProcessor[str](
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                result_processor=processor,
                max_workers=2,
            ) as processor,
        ):
            processor.add_items([])
        assert len(batches) == 0

    def test_iterable_processor_result_processor_error(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test that exceptions in result_processor are handled and logged"""
        url = "https://test.com/api"
        items = [{"id": i} for i in range(2)]
        error_logged = []

        def processor(batch: BatchResult[str]) -> None:
            raise RuntimeError("Test error in result processor")

        class DummyConsole:
            def print(self, msg):
                error_logged.append(msg)

        with (
            responses.RequestsMock() as rsps,
            HTTPBatchProcessor[str](
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                result_processor=processor,
                max_workers=1,
                console=DummyConsole(),
            ) as processor,
        ):
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 1}, {"id": 2}]},
            )
            processor.add_items(items)
        assert any("Error processing result" in str(msg) for msg in error_logged)

    def test_iterable_processor_shutdown_no_items(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test that threads shut down cleanly when no items are added"""
        url = "https://test.com/api"
        batches: list[BatchResult[str]] = []

        def processor(batch: BatchResult[str]) -> None:
            batches.append(batch)

        with HTTPBatchProcessor[str](
            endpoint_url=url,
            config=toolkit_config,
            as_id=lambda item: item["id"],
            result_processor=processor,
            max_workers=2,
        ) as processor:
            pass  # No items added
        assert len(batches) == 0

    def test_iterable_processor_thread_safety(self, toolkit_config: ToolkitClientConfig) -> None:
        """Test adding items from multiple threads is processed correctly"""
        url = "https://test.com/api"
        items = [{"id": i} for i in range(10)]
        batches: list[BatchResult[str]] = []

        def processor(batch: BatchResult[str]) -> None:
            batches.append(batch)

        with (
            responses.RequestsMock() as rsps,
            HTTPBatchProcessor[str](
                endpoint_url=url,
                config=toolkit_config,
                as_id=lambda item: item["id"],
                result_processor=processor,
                max_workers=2,
                batch_size=2,
            ) as processor,
        ):
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 1}, {"id": 2}]},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 3}, {"id": 4}]},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 5}, {"id": 6}]},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 7}, {"id": 8}]},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"id": 9}, {"id": 10}]},
            )

            def add_items_thread(start, end):
                processor.add_items(items[start:end])

            threads = [threading.Thread(target=add_items_thread, args=(i, i + 2)) for i in range(0, 10, 2)]
            for t in threads:
                t.start()
            for t in threads:
                t.join()
        assert len(batches) == 5



from collections.abc import Callable
from typing import Any

from cognite.client.data_classes import Asset, Event, FileMetadata, Sequence, TimeSeries
from cognite.client.data_classes._base import CogniteResource
from cognite.client.data_classes.data_modeling import MappedProperty, NodeApply, NodeId
from cognite.client.data_classes.data_modeling.instances import NodeOrEdgeData, PropertyValueWrite
from cognite.client.data_classes.data_modeling.views import ViewProperty

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId, ViewSource
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    asset_centric_convert_to_primary_property,
    convert_to_primary_property,
)
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentric

from .base import T_AssetCentricResource
from .data_model import INSTANCE_SOURCE_VIEW_ID
from .issues import ConversionIssue, FailedConversion, InvalidPropertyDataType

# These properties are not expected to be mapped to the instance.
_RESERVED_ASSET_CENTRIC_PROPERTIES = frozenset({"metadata", "externalId", "dataSetId", "parentId", "id"})


def asset_centric_to_dm(
    resource: T_AssetCentricResource,
    instance_id: NodeId,
    view_source: ViewSource,
    view_properties: dict[str, ViewProperty],
) -> tuple[NodeApply, ConversionIssue]:
    """Convert an asset-centric resource to a data model instance.

    Args:
        resource (T_AssetCentricResource): The asset-centric resource to convert.
        instance_id (NodeId): The ID of the instance to create or update.
        view_source (ViewSource): The view source defining how to map the resource to the data model.
        view_properties (dict[str, ViewProperty]): The defined properties referenced in the view source mapping.

    Returns:
        tuple[NodeApply, ConversionIssue]: A tuple containing the converted NodeApply and any ConversionIssue encountered.
    """
    resource_type = _lookup_resource_type(type(resource))
    dumped = resource.dump()
    available_properties = (
        set(dumped.keys()) | {f"metadata.{key}" for key in resource.metadata or {}}
    ) - _RESERVED_ASSET_CENTRIC_PROPERTIES
    expected_properties = set(view_source.mapping.to_property_id.keys()) | {
        f"metadata.{key}" for key in (view_source.mapping.metadata_to_property_id or {}).keys()
    }

    issue = ConversionIssue(
        asset_centric_id=AssetCentricId(resource_type, id_=resource.id),
        instance_id=instance_id,
        ignored_asset_centric_properties=sorted(available_properties - expected_properties),
    )

    properties = _create_properties(
        dumped,
        issue,
        conversion=lambda value, prop_id, dm_prop: asset_centric_convert_to_primary_property(
            value,
            dm_prop.type,
            dm_prop.nullable,
            destination_container_property=(dm_prop.container, dm_prop.container_property_identifier),
            source_property=(resource_type, prop_id),
        ),
        view_properties=view_properties,
        asset_centric_to_instance=view_source.mapping.to_property_id,
    )
    if view_source.mapping.metadata_to_property_id is not None:
        metadata_properties = _create_properties(
            resource.metadata or {},
            issue,
            conversion=lambda value, prop_id, dm_prop: convert_to_primary_property(
                value,
                dm_prop.type,
                dm_prop.nullable,
            ),
            view_properties=view_properties,
            asset_centric_to_instance=view_source.mapping.metadata_to_property_id,
            source_prefix="metadata.",
        )
        for key, value in metadata_properties.items():
            if key not in properties:
                properties[key] = value
            else:
                issue.ignored_asset_centric_properties.append(f"metadata.{key}")

    sources: list[NodeOrEdgeData] = []
    if properties:
        sources.append(NodeOrEdgeData(source=view_source.view_id, properties=properties))
    sources.append(
        NodeOrEdgeData(
            source=INSTANCE_SOURCE_VIEW_ID,
            properties={
                "resourceType": resource_type,
                "id": resource.id,
                "dataSetId": resource.data_set_id,
                "classicExternalId": resource.external_id,
            },
        )
    )

    node = NodeApply(
        space=instance_id.space,
        external_id=instance_id.external_id,
        sources=sources,
    )

    return node, issue


def _lookup_resource_type(resource_type: type[T_AssetCentricResource]) -> AssetCentric:
    resource_type_map: dict[type[CogniteResource], AssetCentric] = {
        Asset: "asset",
        FileMetadata: "file",
        Event: "event",
        TimeSeries: "timeseries",
        Sequence: "sequence",
    }
    try:
        return resource_type_map[resource_type]
    except KeyError as e:
        raise ValueError(f"Unsupported resource type: {resource_type}") from e


def _create_properties(
    dumped: dict[str, Any],
    issue: ConversionIssue,
    conversion: Callable[[Any, str, MappedProperty], PropertyValueWrite],
    view_properties: dict[str, ViewProperty],
    asset_centric_to_instance: dict[str, str],
    source_prefix: str = "",
) -> dict[str, PropertyValueWrite]:
    properties: dict[str, PropertyValueWrite] = {}
    for prop_id, dm_prop_id in asset_centric_to_instance.items():
        if prop_id not in dumped:
            issue.missing_asset_centric_properties.append(f"{source_prefix}{prop_id}")
            continue
        if dm_prop_id not in view_properties:
            issue.missing_instance_properties.append(dm_prop_id)
            continue
        dm_prop = view_properties[dm_prop_id]
        if not isinstance(dm_prop, MappedProperty):
            issue.invalid_instance_property_types.append(
                InvalidPropertyDataType(property_id=dm_prop_id, expected_type=MappedProperty.__name__)
            )
            continue
        try:
            value = conversion(dumped[prop_id], prop_id, dm_prop)
        except (ValueError, TypeError) as e:
            issue.failed_conversions.append(FailedConversion(property_id=prop_id, value=dumped[prop_id], error=str(e)))
            continue
        properties[dm_prop_id] = value
    return properties



from pathlib import Path

import pytest
from cognite.client.data_classes import (
    Asset,
    AssetList,
    AssetWriteList,
    CountAggregate,
    DataSet,
    DataSetList,
    FileMetadataList,
    FileMetadataWriteList,
    LabelDefinition,
    LabelDefinitionList,
)

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import DownloadCommand, UploadCommand
from cognite_toolkit._cdf_tk.storageio import AssetIO, FileMetadataIO
from cognite_toolkit._cdf_tk.storageio._selectors import AssetSubtreeSelector, DataSetSelector
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


@pytest.fixture()
def some_asset_data() -> AssetList:
    """Fixture to provide a sample AssetList for testing."""
    return AssetList(
        [
            Asset(
                external_id=f"asset_{i}",
                name=f"Asset {i}",
                description=f"Description for asset {i}",
                root_id=123,
                source="test_source",
                labels=["my_label"],
                data_set_id=1234,
            )
            for i in range(100)
        ]
    )


class TestAssetIO:
    def test_download_upload(self, some_asset_data: AssetList) -> None:
        selector = AssetSubtreeSelector(hierarchy="test_hierarchy")
        with monkeypatch_toolkit_client() as client:
            client.assets.return_value = chunker(some_asset_data, 10)
            client.assets.aggregate_count.return_value = 100
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234

            io = AssetIO(client)

            assert io.count(selector) == 100

            source = io.download_iterable(selector)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for chunk in source:
                json_chunk = io.data_to_json_chunk(chunk)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                    assert "dataSetExternalId" in item
                    assert item["dataSetExternalId"] == "test_data_set"
                json_chunks.append(json_chunk)

            data_chunks = (io.json_chunk_to_data(chunk) for chunk in json_chunks)
            for data_chunk in data_chunks:
                io.upload_items(data_chunk, selector)

            assert client.assets.upsert.call_count == 10
            uploaded_assets = AssetWriteList([])
            for call in client.assets.upsert.call_args_list:
                uploaded_assets.extend(call[0][0])

            assert uploaded_assets.dump() == some_asset_data.as_write().dump()

    def test_download_upload_command(self, some_asset_data: AssetList, tmp_path: Path) -> None:
        selector = AssetSubtreeSelector(hierarchy="test_hierarchy")
        with monkeypatch_toolkit_client() as client:
            client.assets.return_value = [some_asset_data]
            client.assets.aggregate_count.return_value = 100
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234
            client.lookup.assets.external_id.return_value = "test_hierarchy"
            client.lookup.assets.id.return_value = 123
            client.data_sets.retrieve_multiple.return_value = DataSetList(
                [DataSet(id=1234, external_id="test_data_set")]
            )
            client.labels.retrieve.return_value = LabelDefinitionList(
                [LabelDefinition(external_id="my_label", name="my_label")]
            )

            io = AssetIO(client)

            download_command = DownloadCommand(silent=True, skip_tracking=True)
            upload_command = UploadCommand(silent=True, skip_tracking=True)

            download_command.download(
                selectors=[selector],
                io=io,
                output_dir=tmp_path,
                verbose=False,
                file_format=".csv",
                compression="none",
                limit=100,
            )

            upload_command.upload(
                io=io,
                input_dir=tmp_path / io.folder_name,
                ensure_configurations=True,
                dry_run=False,
                verbose=False,
            )

            assert client.assets.upsert.call_count == 1
            args, _ = client.assets.upsert.call_args
            assert len(args) == 1
            uploaded_assets = args[0]
            assert isinstance(uploaded_assets, AssetWriteList)
            assert [asset.dump() for asset in uploaded_assets] == [asset.as_write().dump() for asset in some_asset_data]


@pytest.fixture()
def some_filemetadata_data() -> FileMetadataList:
    """Fixture to provide a sample FileMetadataList for testing."""
    from cognite.client.data_classes import FileMetadata

    return FileMetadataList(
        [
            FileMetadata(
                external_id=f"file_{i}",
                name=f"File {i}",
                directory="/test/dir",
                mime_type="text/plain",
                data_set_id=1234,
                asset_ids=[123],
                source="test_source",
            )
            for i in range(50)
        ]
    )


class TestFileMetadataIO:
    def test_download_upload(self, some_filemetadata_data: FileMetadataList) -> None:
        selector = DataSetSelector(data_set_external_id="DataSetSelector")
        with monkeypatch_toolkit_client() as client:
            client.files.return_value = chunker(some_filemetadata_data, 10)
            client.files.aggregate.return_value = [CountAggregate(50)]
            client.lookup.data_sets.external_id.return_value = "test_data_set"
            client.lookup.data_sets.id.return_value = 1234
            client.lookup.assets.external_id.return_value = ["test_hierarchy"]
            client.lookup.assets.id.return_value = [123]

            io = FileMetadataIO(client)

            assert io.count(selector) == 50

            source = io.download_iterable(selector)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for chunk in source:
                json_chunk = io.data_to_json_chunk(chunk)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                    assert "dataSetExternalId" in item
                    assert item["dataSetExternalId"] == "test_data_set"
                json_chunks.append(json_chunk)

            data_chunks = (io.json_chunk_to_data(chunk) for chunk in json_chunks)
            for data_chunk in data_chunks:
                io.upload_items(data_chunk, selector)

            assert client.files.create.call_count == len(some_filemetadata_data)
            uploaded_files = FileMetadataWriteList([])
            for call in client.files.create.call_args_list:
                uploaded_files.append(call[0][0])

            assert uploaded_files.dump() == some_filemetadata_data.as_write().dump()



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import EventYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_event_test_cases() -> Iterable:
    yield pytest.param(
        {"type": "Workorder"}, {"Missing required field: 'externalId'"}, id="Missing required field: externalId"
    )
    yield pytest.param(
        {"externalId": "1230098u-12907903", "assetExternalIds": [123, 456], "startTime": "2023-10-01T00:00:00Z"},
        {
            "In assetExternalIds[1] input should be a valid string. Got 123 of type int. "
            "Hint: Use double quotes to force string.",
            "In assetExternalIds[2] input should be a valid string. Got 456 of type int. "
            "Hint: Use double quotes to force string.",
            "In field startTime input should be a valid integer. Got '2023-10-01T00:00:00Z' of type str.",
        },
        id="Invalid data type in assetExternalIds",
    )


class TestEventYAML:
    @pytest.mark.parametrize("data", list(find_resources("Event")))
    def test_load_valid_asset(self, data: dict[str, object]) -> None:
        loaded = EventYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_event_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, EventYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from pathlib import Path

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.commands import DownloadCommand
from cognite_toolkit._cdf_tk.storageio import RawIO
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file
from cognite_toolkit._cdf_tk.utils.fileio import NDJsonReader
from tests.test_integration.constants import TIMESERIES_COUNT, TIMESERIES_TABLE


class TestDownloadCommand:
    def test_download_raw_table(self, toolkit_client: ToolkitClient, aggregator_raw_db: str, tmp_path: Path) -> None:
        cmd = DownloadCommand(silent=True, skip_tracking=True)
        table = RawTable(db_name=aggregator_raw_db, table_name=TIMESERIES_TABLE)
        cmd.download(
            [table],
            RawIO(toolkit_client),
            output_dir=tmp_path,
            verbose=False,
            file_format=".ndjson",
            compression="none",
            limit=None,
        )
        downloaded_files = list(tmp_path.rglob("*.ndjson"))
        assert len(downloaded_files) == 1, "Expected exactly one file to be downloaded."
        chunks = list(NDJsonReader(downloaded_files[0]).read_chunks())
        assert len(chunks) == TIMESERIES_COUNT, f"Expected {TIMESERIES_COUNT} chunks, got {len(chunks)}."
        config_files = list(tmp_path.rglob("*.yaml"))
        assert len(config_files) == 1, "Expected exactly one configuration file to be created."
        dumped = read_yaml_file(config_files[0], "dict")
        assert dumped == table.dump()



import random
import time
from collections.abc import Iterable
from pathlib import Path

import pytest
from cognite.client.data_classes import FileMetadataList, FileMetadataWrite
from cognite.client.data_classes.data_modeling import NodeId, Space
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.extended_timeseries import ExtendedTimeSeriesList
from cognite_toolkit._cdf_tk.commands import MigrateFilesCommand


@pytest.fixture()
def three_files_with_content(
    toolkit_client_with_pending_ids: ToolkitClient, toolkit_space: Space
) -> Iterable[FileMetadataList]:
    client = toolkit_client_with_pending_ids
    space = toolkit_space.space
    files = FileMetadataList([])
    for i in range(3):
        file = FileMetadataWrite(
            external_id=f"toolkit_test_migration_{i}_{random.randint(0, 10_000)!s}",
            name=f"toolkit_test_migration_{i}",
        )
        files.append(file)
    output = client.files.retrieve_multiple(external_ids=files.as_external_ids(), ignore_unknown_ids=True)
    if output:
        try:
            client.files.delete(external_id=output.as_external_ids(), ignore_unknown_ids=True)
        except CogniteAPIError:
            client.data_modeling.instances.delete([NodeId(space, ts.external_id) for ts in output])
    created = FileMetadataList([])
    for file in files:
        uploaded = client.files.upload_bytes(b"test content", **file.dump(camel_case=False))
        created.append(uploaded)

    yield created

    # Cleanup after test
    deleted = client.data_modeling.instances.delete([NodeId(space, file.external_id) for file in created])
    if deleted.nodes:
        return
    client.files.delete(external_id=created.as_external_ids(), ignore_unknown_ids=True)


@pytest.mark.skip(
    "This is not yet enabled in the staging cluster that Toolkit uses for testing. Only runs a dev cluster."
)
class TestMigrateFilesCommand:
    # This tests uses instances.apply_fast() which uses up to 4 workers for writing instances,
    # when this is used in parallel with other tests that uses instances.apply() then we get 5 workers in total,
    # which will trigger a 429 error.
    @pytest.mark.usefixtures("max_two_workers")
    def test_migrate_files_command(
        self,
        toolkit_client_with_pending_ids: ToolkitClient,
        three_files_with_content: ExtendedTimeSeriesList,
        toolkit_space: Space,
        tmp_path: Path,
    ) -> None:
        client = toolkit_client_with_pending_ids
        space = toolkit_space.space

        input_file = tmp_path / "files_migration.csv"
        with input_file.open("w", encoding="utf-8") as f:
            f.write(
                "id,dataSetId,space,externalId\n"
                + "\n".join(
                    f"{ts.id},{ts.data_set_id if ts.data_set_id else ''},{space},{ts.external_id}"
                    for ts in three_files_with_content
                )
                + "\n"
            )

        cmd = MigrateFilesCommand(skip_tracking=True, silent=True)
        cmd.migrate_files(
            client=client,
            mapping_file=input_file,
            dry_run=False,
            verbose=False,
            auto_yes=True,
        )
        # Wait for syncer
        time.sleep(5)

        migrated_files = client.files.retrieve_multiple(external_ids=three_files_with_content.as_external_ids())

        missing_node_id = [ts.external_id for ts in migrated_files if ts.instance_id is None]
        assert not missing_node_id, f"Some files are missing NodeId: {missing_node_id}"

        node_ids = [ts.instance_id for ts in migrated_files]
        for node_id in node_ids:
            content = client.files.download_bytes(instance_id=node_id)
            assert content == b"test content", f"Content of file {node_id} does not match expected content."



import gzip
import random
import socket
import sys
import time
from collections import deque
from collections.abc import MutableMapping, Sequence, Set
from typing import Literal

import requests
import urllib3
from cognite.client import global_config
from cognite.client.utils import _json
from requests.adapters import HTTPAdapter
from requests.structures import CaseInsensitiveDict
from urllib3.util.retry import Retry

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils.auxiliary import get_current_toolkit_version, get_user_agent
from cognite_toolkit._cdf_tk.utils.http_client._data_classes import (
    BodyRequest,
    FailedRequestMessage,
    HTTPMessage,
    ItemsRequest,
    ParamRequest,
    RequestMessage,
    ResponseMessage,
)
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class HTTPClient:
    """An HTTP client.

    This class handles rate limiting, retries, and error handling for HTTP requests.

    Args:
        config (ToolkitClientConfig): Configuration for the Toolkit client.
        pool_connections (int): The number of connection pools to cache. Default is 10.
        pool_maxsize (int): The maximum number of connections to save in the pool. Default
            is 20.
        max_retries (int): The maximum number of retries for a request. Default is 10.
        retry_status_codes (frozenset[int]): HTTP status codes that should trigger a retry.
            Default is {408, 429, 502, 503, 504}.
        split_items_status_codes (frozenset[int]): In the case of ItemRequest with multiple
            items, these status codes will trigger splitting the request into smaller batches.

    """

    def __init__(
        self,
        config: ToolkitClientConfig,
        max_retries: int = 10,
        pool_connections: int = 10,
        pool_maxsize: int = 20,
        retry_status_codes: Set[int] = frozenset({408, 429, 502, 503, 504}),
        split_items_status_codes: Set[int] = frozenset({400, 408, 409, 422, 502, 503, 504}),
    ):
        self.config = config
        self._max_retries = max_retries
        self._pool_connections = pool_connections
        self._pool_maxsize = pool_maxsize
        self._retry_status_codes = retry_status_codes
        self._split_items_status_codes = split_items_status_codes

        # Thread-safe session for connection pooling
        self.session = self._create_thread_safe_session()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: object | None
    ) -> Literal[False]:
        """Close the session when exiting the context."""
        self.session.close()
        return False  # Do not suppress exceptions

    def request(self, message: RequestMessage) -> Sequence[HTTPMessage]:
        """Send an HTTP request and return the response.

        Args:
            message (RequestMessage): The request message to send.

        Returns:
            Sequence[HTTPMessage]: The response message(s). This can also
                include RequestMessage(s) to be retried.
        """
        if isinstance(message, ItemsRequest) and message.tracker and message.tracker.limit_reached():
            error_msg = (
                f"Aborting further splitting of requests after {message.tracker.failed_split_count} failed attempts."
            )
            return message.create_failed_request(error_msg)
        try:
            response = self._make_request(message)
            results = self._handle_response(response, message)
        except Exception as e:
            results = self._handle_error(e, message)
        return results

    def request_with_retries(self, message: RequestMessage) -> Sequence[ResponseMessage | FailedRequestMessage]:
        """Send an HTTP request and handle retries.

        This method will keep retrying the request until it either succeeds or
        exhausts the maximum number of retries.

        Note this method will use the current thread to process all request, thus
        it is blocking.

        Args:
            message (RequestMessage): The request message to send.

        Returns:
            Sequence[ResponseMessage | FailedRequestMessage]: The final response
                messages, which can be either successful responses or failed requests.
        """
        if message.total_attempts > 0:
            raise RuntimeError(f"RequestMessage has already been attempted {message.total_attempts} times.")
        pending_requests: deque[RequestMessage] = deque()
        pending_requests.append(message)
        final_responses: list[ResponseMessage | FailedRequestMessage] = []

        while pending_requests:
            current_request = pending_requests.popleft()
            results = self.request(current_request)

            for result in results:
                if isinstance(result, RequestMessage):
                    pending_requests.append(result)
                elif isinstance(result, ResponseMessage | FailedRequestMessage):
                    final_responses.append(result)
                else:
                    raise TypeError(f"Unexpected result type: {type(result)}")

        return final_responses

    def _create_thread_safe_session(self) -> requests.Session:
        session = requests.Session()
        adapter = HTTPAdapter(
            pool_connections=self._pool_connections,
            pool_maxsize=self._pool_maxsize,
            max_retries=Retry(total=0),  # We handle retries manually
        )
        session.mount("https://", adapter)
        return session

    def _create_headers(self, api_version: str | None = None) -> MutableMapping[str, str]:
        headers: MutableMapping[str, str] = CaseInsensitiveDict()
        headers.update(requests.utils.default_headers())
        auth_name, auth_value = self.config.credentials.authorization_header()
        headers[auth_name] = auth_value
        headers["content-type"] = "application/json"
        headers["accept"] = "application/json"
        headers["x-cdp-sdk"] = f"CogniteToolkit:{get_current_toolkit_version()}"
        headers["x-cdp-app"] = self.config.client_name
        headers["cdf-version"] = api_version or self.config.api_subversion
        if "User-Agent" in headers:
            headers["User-Agent"] += f" {get_user_agent()}"
        else:
            headers["User-Agent"] = get_user_agent()
        if not global_config.disable_gzip:
            headers["Content-Encoding"] = "gzip"
        return headers

    @staticmethod
    def _prepare_payload(item: BodyRequest) -> str | bytes:
        """
        Prepare the payload for the HTTP request.
        This method should be overridden in subclasses to customize the payload format.
        """
        data: str | bytes
        try:
            data = _json.dumps(item.body(), allow_nan=False)
        except ValueError as e:
            # A lot of work to give a more human friendly error message when nans and infs are present:
            msg = "Out of range float values are not JSON compliant"
            if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
                raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                    e.__traceback__
                ) from None
            raise

        if not global_config.disable_gzip:
            data = gzip.compress(data.encode())
        return data

    def _make_request(self, item: RequestMessage) -> requests.Response:
        headers = self._create_headers(item.api_version)
        params: dict[str, str] | None = None
        if isinstance(item, ParamRequest):
            params = item.parameters
        data: str | bytes | None = None
        if isinstance(item, BodyRequest):
            data = self._prepare_payload(item)
        return self.session.request(
            method=item.method,
            url=item.endpoint_url,
            data=data,
            headers=headers,
            params=params,
            timeout=self.config.timeout,
            allow_redirects=False,
        )

    def _handle_response(
        self,
        response: requests.Response,
        request: RequestMessage,
    ) -> Sequence[HTTPMessage]:
        try:
            body = response.json()
        except ValueError as e:
            return request.create_responses(response, error_message=f"Invalid JSON response: {e!s}")

        if 200 <= response.status_code < 300:
            return request.create_responses(response, body)
        elif (
            isinstance(request, ItemsRequest)
            and len(request.items) > 1
            and response.status_code in self._split_items_status_codes
        ):
            # 4XX: Status there is at least one item that is invalid, split the batch to get all valid items processed
            # 5xx: Server error, split to reduce the number of items in each request, and count as a status attempt
            status_attempts = request.status_attempt
            if 500 <= response.status_code < 600:
                status_attempts += 1
            splits = request.split(status_attempts=status_attempts)
            if splits[0].tracker and splits[0].tracker.limit_reached():
                return request.create_responses(response, body, self._get_error_message(body, response.text))
            return splits
        elif request.status_attempt < self._max_retries and response.status_code in self._retry_status_codes:
            request.status_attempt += 1
            time.sleep(self._backoff_time(request.total_attempts))
            return [request]
        else:
            # Permanent failure
            return request.create_responses(response, body, self._get_error_message(body, response.text))

    @staticmethod
    def _get_error_message(body: JsonVal, default: str) -> str:
        error = default
        if not isinstance(body, dict):
            return error
        if "error" not in body:
            return error
        error_nested = body["error"]
        if isinstance(error_nested, str):
            return error_nested
        if isinstance(error_nested, dict) and "message" in error_nested and isinstance(error_nested["message"], str):
            return error_nested["message"]
        return error

    @staticmethod
    def _backoff_time(attempts: int) -> float:
        backoff_time = 0.5 * (2**attempts)
        return min(backoff_time, global_config.max_retry_backoff) * random.uniform(0, 1.0)

    def _handle_error(
        self,
        e: Exception,
        request: RequestMessage,
    ) -> Sequence[HTTPMessage]:
        if self._any_exception_in_context_isinstance(
            e, (socket.timeout, urllib3.exceptions.ReadTimeoutError, requests.exceptions.ReadTimeout)
        ):
            error_type = "read"
            request.read_attempt += 1
            attempts = request.read_attempt
        elif self._any_exception_in_context_isinstance(
            e,
            (
                ConnectionError,
                urllib3.exceptions.ConnectionError,
                urllib3.exceptions.ConnectTimeoutError,
                requests.exceptions.ConnectionError,
            ),
        ):
            error_type = "connect"
            request.connect_attempt += 1
            attempts = request.connect_attempt
        else:
            error_msg = f"Unexpected exception: {e!s}"
            return request.create_failed_request(error_msg)

        if attempts <= self._max_retries:
            time.sleep(self._backoff_time(request.total_attempts))
            return [request]
        else:
            # We have already incremented the attempt count, so we subtract 1 here
            error_msg = f"RequestException after {request.total_attempts - 1} attempts ({error_type} error): {e!s}"

            return request.create_failed_request(error_msg)

    @classmethod
    def _any_exception_in_context_isinstance(
        cls, exc: BaseException, exc_types: tuple[type[BaseException], ...] | type[BaseException]
    ) -> bool:
        """requests does not use the "raise ... from ..." syntax, so we need to access the underlying exceptions using
        the __context__ attribute.
        """
        if isinstance(exc, exc_types):
            return True
        if exc.__context__ is None:
            return False
        return cls._any_exception_in_context_isinstance(exc.__context__, exc_types)



from collections.abc import Iterable, Mapping
from pathlib import Path
from types import MappingProxyType
from typing import ClassVar

from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling import Edge, EdgeApply, Node, NodeApply
from cognite.client.utils._identifier import InstanceId
from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.instances import InstanceApplyList, InstanceList
from cognite_toolkit._cdf_tk.utils.cdf import iterate_instances
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.fileio import SchemaColumn
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import StorageIOConfig, TableStorageIO
from ._selectors import InstanceFileSelector, InstanceSelector, InstanceViewSelector


class InstanceIO(TableStorageIO[InstanceId, InstanceSelector, InstanceApplyList, InstanceList]):
    folder_name = "instances"
    kind = "Instances"
    display_name = "Instances"
    supported_download_formats = frozenset({".parquet", ".csv", ".ndjson"})
    supported_compressions = frozenset({".gz"})
    supported_read_formats = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    chunk_size = 1000
    UPLOAD_ENDPOINT = "/models/instances"
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = MappingProxyType({"autoCreateDirectRelations": True})

    def as_id(self, item: dict[str, JsonVal] | object) -> InstanceId:
        if isinstance(item, dict) and isinstance(item.get("space"), str) and isinstance(item.get("externalId"), str):
            # MyPy checked above.
            return InstanceId(space=item["space"], external_id=item["externalId"])  # type: ignore[arg-type]
        if isinstance(item, InstanceId):
            return item
        elif isinstance(item, Node | Edge | NodeApply | EdgeApply):
            return item.as_id()
        raise TypeError(f"Cannot extract ID from item of type {type(item).__name__!r}")

    def download_iterable(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[InstanceList]:
        if isinstance(selector, InstanceViewSelector):
            chunk = InstanceList([])
            total = 0
            for instance in iterate_instances(
                client=self.client,
                source=selector.view,
                instance_type=selector.instance_type,
                space=list(selector.instance_spaces) if selector.instance_spaces else None,
            ):
                if limit is not None and total >= limit:
                    break
                total += 1
                chunk.append(instance)
                if len(chunk) >= self.chunk_size:
                    yield chunk
                    chunk = InstanceList([])
            if chunk:
                yield chunk
        elif isinstance(selector, InstanceFileSelector):
            for node_chunk in chunker_sequence(selector.node_ids, self.chunk_size):
                yield InstanceList(self.client.data_modeling.instances.retrieve(nodes=node_chunk).nodes)
            for edge_chunk in chunker_sequence(selector.edge_ids, self.chunk_size):
                yield InstanceList(self.client.data_modeling.instances.retrieve(edges=edge_chunk).edges)
        else:
            raise NotImplementedError()

    def download_ids(self, selector: InstanceSelector, limit: int | None = None) -> Iterable[list[InstanceId]]:
        if isinstance(selector, InstanceFileSelector) and selector.validate is False:
            instances_to_yield = selector.instance_ids
            if limit is not None:
                instances_to_yield = instances_to_yield[:limit]
            yield from chunker_sequence(instances_to_yield, self.chunk_size)
        else:
            yield from ([instance.as_id() for instance in chunk] for chunk in self.download_iterable(selector, limit))  # type: ignore[attr-defined]

    def count(self, selector: InstanceSelector) -> int | None:
        if isinstance(selector, InstanceViewSelector):
            result = self.client.data_modeling.instances.aggregate(
                view=selector.view,
                aggregates=Count("externalId"),
                instance_type=selector.instance_type,
                space=list(selector.instance_spaces) if selector.instance_spaces else None,
            )
            return int(result.value or 0)
        elif isinstance(selector, InstanceFileSelector):
            return len(selector.instance_ids)
        raise NotImplementedError()

    def upload_items(self, data_chunk: InstanceApplyList, selector: InstanceSelector) -> None:
        raise NotImplementedError()

    def data_to_json_chunk(self, data_chunk: InstanceList) -> list[dict[str, JsonVal]]:
        raise NotImplementedError()

    def json_chunk_to_data(self, data_chunk: list[dict[str, JsonVal]]) -> InstanceApplyList:
        raise NotImplementedError()

    def configurations(self, selector: InstanceSelector) -> Iterable[StorageIOConfig]:
        raise NotImplementedError()

    def load_selector(self, datafile: Path) -> InstanceSelector:
        raise NotImplementedError()

    def ensure_configurations(self, selector: InstanceSelector, console: Console | None = None) -> None:
        raise NotImplementedError()

    def get_schema(self, selector: InstanceSelector) -> list[SchemaColumn]:
        raise NotImplementedError()



from unittest.mock import MagicMock

import pytest
from cognite.client._api.iam import IAMAPI, ComparableCapability
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    DataSetScope,
    ProjectCapabilityList,
    TimeSeriesAcl,
)
from cognite.client.data_classes.data_modeling import DataModel, DataModelList
from cognite.client.data_classes.data_modeling.statistics import InstanceStatistics, ProjectStatistics

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands._migrate.base import BaseMigrateCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_model import (
    INSTANCE_SOURCE_VIEW_ID,
    MODEL_ID,
    RESOURCE_VIEW_MAPPING_VIEW_ID,
)
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError, ToolkitMigrationError, ToolkitValueError


class DummyMigrationCommand(BaseMigrateCommand):
    """A dummy command for testing purposes."""

    def source_acl(self, data_set_id: list[int]) -> TimeSeriesAcl:
        return TimeSeriesAcl(
            actions=[TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
            scope=DataSetScope(data_set_id),
        )


class TestBaseCommand:
    def test_validate_access(self) -> None:
        cmd = DummyMigrationCommand(silent=True)
        expected_missing = [
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Read,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID(["my_instance_space"]),
            ),
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID(["my_instance_space"]),
            ),
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write_Properties,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID(["my_instance_space"]),
            ),
            DataModelsAcl(actions=[DataModelsAcl.Action.Read], scope=DataModelsAcl.Scope.SpaceID(["dummy_space"])),
            TimeSeriesAcl(actions=[TimeSeriesAcl.Action.Read], scope=DataSetScope([123])),
            TimeSeriesAcl(actions=[TimeSeriesAcl.Action.Write], scope=DataSetScope([123])),
        ]

        with monkeypatch_toolkit_client() as client:
            existing_capabilities = ProjectCapabilityList([], cognite_client=client)

            def verify_capabilities(desired_capabilities: ComparableCapability) -> list[Capability]:
                return IAMAPI.compare_capabilities(
                    existing_capabilities,
                    desired_capabilities,
                )

            client.iam.verify_capabilities = verify_capabilities
            with pytest.raises(AuthenticationError) as exc_info:
                cmd.validate_access(client, ["my_instance_space"], ["dummy_space"], [123])
        error = exc_info.value
        assert isinstance(error, AuthenticationError)
        assert "Missing required capabilities" in error.args[0]
        assert error.args[1] == expected_missing

    def test_validate_migration_model_available(self) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.data_models.retrieve.return_value = DataModelList([])
            with pytest.raises(ToolkitMigrationError):
                BaseMigrateCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_multiple_models(self) -> None:
        """Test that multiple models raises an error."""
        with monkeypatch_toolkit_client() as client:
            # Create mock models with the expected MODEL_ID
            model1 = MagicMock(spec=DataModel)
            model1.as_id.return_value = MODEL_ID
            model2 = MagicMock(spec=DataModel)
            model2.as_id.return_value = MODEL_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model1, model2])

            with pytest.raises(ToolkitMigrationError) as exc_info:
                BaseMigrateCommand.validate_migration_model_available(client)

            assert "Multiple migration models" in str(exc_info.value)

    def test_validate_migration_model_available_missing_views(self) -> None:
        """Test that a model with missing views raises an error."""
        with monkeypatch_toolkit_client() as client:
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has views but missing the required ones
            model.views = [INSTANCE_SOURCE_VIEW_ID]  # Missing VIEW_SOURCE_VIEW_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            with pytest.raises(ToolkitMigrationError, match=r"Invalid migration model. Missing views"):
                BaseMigrateCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_success(self) -> None:
        """Test that a valid model with all required views succeeds."""
        with monkeypatch_toolkit_client() as client:
            # Mocking the migration Model to get a response format of the model.
            # An alternative would be to write a conversion of write -> read format of the model
            # which is a significant amount of logic.
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has all required views
            model.views = [INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID]

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            # Should not raise any exception
            BaseMigrateCommand.validate_migration_model_available(client)

            client.data_modeling.data_models.retrieve.assert_called_once_with([MODEL_ID], inline_views=False)

    def test_validate_available_capacity_missing_capacity(self) -> None:
        cmd = DummyMigrationCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            stats = MagicMock(spec=ProjectStatistics)
            stats.instances = InstanceStatistics(
                nodes=1000,
                edges=0,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
                instances_limit=1500,
                soft_deleted_instances_limit=10_000,
                instances=1000,
                soft_deleted_instances=0,
            )
            client.data_modeling.statistics.project.return_value = stats
            with pytest.raises(ToolkitValueError) as exc_info:
                cmd.validate_available_capacity(client, 10_000)

        assert "Cannot proceed with migration" in str(exc_info.value)

    def test_validate_available_capacity_sufficient_capacity(self) -> None:
        cmd = DummyMigrationCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            stats = MagicMock(spec=ProjectStatistics)
            stats.instances = InstanceStatistics(
                nodes=1000,
                edges=0,
                soft_deleted_edges=0,
                soft_deleted_nodes=0,
                instances_limit=5_000_000,
                soft_deleted_instances_limit=100_000_000,
                instances=1000,
                soft_deleted_instances=0,
            )
            client.data_modeling.statistics.project.return_value = stats
            cmd.validate_available_capacity(client, 10_000)



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import AssetYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_asset_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "Asset 1"}, {"Missing required field: 'externalId'"}, id="Missing required field: externalId"
    )
    yield pytest.param(
        {"externalId": "asset_1", "name": "Asset 1", "parentId": 123},
        {"Unused field: 'parentId'"},
        id="Unused field: parentId",
    )


class TestAssetYAML:
    @pytest.mark.parametrize("data", list(find_resources("Asset")))
    def test_load_valid_asset(self, data: dict[str, object]) -> None:
        loaded = AssetYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_asset_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, AssetYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from pathlib import Path

import pytest
from cognite.client.data_classes.raw import RowWrite

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.cruds import RawTableCRUD
from cognite_toolkit._cdf_tk.storageio import RawIO
from cognite_toolkit._cdf_tk.utils.fileio import NDJsonWriter, Uncompressed


@pytest.fixture
def raw_directory(tmp_path: Path) -> Path:
    """Fixture to create a temporary folder with a sample NDJSON file."""
    folder = tmp_path / RawIO.folder_name
    folder.mkdir(parents=True, exist_ok=True)
    configfile = folder / f"test_table.{RawTableCRUD.kind}.yaml"
    table = RawTable(db_name="test_db", table_name="test_table")
    configfile.write_text(table.dump_yaml())
    with NDJsonWriter(folder, RawIO.kind, Uncompressed) as writer:
        writer.write_chunks(
            [
                RowWrite(
                    key=f"key{i}",
                    columns={
                        "column1": f"value{i}",
                        "column2": i,
                        "column3": i % 2 == 0,
                    },
                ).dump()
                for i in range(1_000)
            ],
            filestem="test_table",
        )

    return folder


class TestUploadCommand:
    def test_upload_raw_rows(self, raw_directory: Path, tmp_path: Path) -> None:
        cmd = UploadCommand(silent=True, skip_tracking=True)
        with monkeypatch_toolkit_client() as client:
            cmd.upload(RawIO(client), raw_directory, ensure_configurations=True, dry_run=False, verbose=False)

            client.raw.rows.insert.assert_called_once()
            _, kwargs = client.raw.rows.insert.call_args
            assert kwargs["db_name"] == "test_db"
            assert kwargs["table_name"] == "test_table"
            inserted_rows = kwargs["row"]
            assert len(inserted_rows) == 1_000
            assert all(isinstance(row, RowWrite) for row in inserted_rows)

            client.raw.databases.create.assert_called_once()
            db_args, _ = client.raw.databases.create.call_args
            assert db_args[0] == ["test_db"]
            client.raw.tables.create.assert_called_once()
            _, table_kwargs = client.raw.tables.create.call_args
            assert table_kwargs["db_name"] == "test_db"
            assert table_kwargs["name"] == ["test_table"]

    def test_upload_raw_rows_dry_run(self, raw_directory: Path) -> None:
        cmd = UploadCommand(silent=True, skip_tracking=True)
        with monkeypatch_toolkit_client() as client:
            cmd.upload(RawIO(client), raw_directory, ensure_configurations=True, dry_run=True, verbose=False)

            client.raw.rows.insert.assert_not_called()
            client.raw.databases.create.assert_not_called()
            client.raw.tables.create.assert_not_called()



from unittest.mock import MagicMock, patch

import pytest
import responses
from cognite.client import global_config
from cognite.client.data_classes import FunctionWrite
from cognite.client.exceptions import CogniteAPIError, CogniteProjectAccessError
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.api.extended_functions import ExtendedFunctionsAPI


class TestExtendedFunctionsAPI:
    def test_create_function_200(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/functions"
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"externalId": "test_function", "name": "test_function", "createdTime": 42}]},
            )
            result = client.functions.create_with_429_retry(fun)

        assert result.external_id == "test_function"
        assert result.name == "test_function"
        assert result.created_time == 42

    def test_create_function_429_succeed(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/functions"
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        console = MagicMock(spec=Console)
        with responses.RequestsMock() as rsps, patch(f"{ExtendedFunctionsAPI.__module__}.time.sleep"):
            for i in range(global_config.max_retries - 1):
                rsps.add(
                    responses.POST, url, status=429, json={"error": "Too many requests"}, headers={"Retry-After": "42"}
                )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"externalId": "test_function", "name": "test_function", "createdTime": 42}]},
            )

            result = client.functions.create_with_429_retry(fun, console=console)
        assert result.external_id == "test_function"
        assert console.print.call_count == global_config.max_retries - 1
        assert console.print.call_args.args[1] == "Rate limit exceeded. Retrying after 42.0 seconds."

    def test_create_function_429_exceed_max_retries(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/functions"
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        console = MagicMock(spec=Console)
        with responses.RequestsMock() as rsps, patch(f"{ExtendedFunctionsAPI.__module__}.time.sleep"):
            for i in range(global_config.max_retries + 1):
                rsps.add(
                    responses.POST, url, status=429, json={"error": "Too many requests"}, headers={"Retry-After": "42"}
                )
            with pytest.raises(CogniteAPIError) as exc_info:
                client.functions.create_with_429_retry(fun, console=console)
        assert console.print.call_count == global_config.max_retries
        assert exc_info.value.message == "Too many requests"
        assert exc_info.value.code == 429
        assert "Rate limit exceeded. Retrying after 42.0 seconds." in console.print.call_args.args[1]

    def test_create_function_429_invalid_retry_after(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/functions"
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        console = MagicMock(spec=Console)

        def mock_uniform(*args, **kwargs):
            return 0

        with (
            responses.RequestsMock() as rsps,
            patch(f"{ExtendedFunctionsAPI.__module__}.time.sleep"),
            patch(f"{ExtendedFunctionsAPI.__module__}.random.uniform", new=mock_uniform),
        ):
            rsps.add(
                responses.POST, url, status=429, json={"error": "Too many requests"}, headers={"Retry-After": "invalid"}
            )
            with pytest.raises(CogniteAPIError) as exc_info:
                client.functions.create_with_429_retry(fun, console=console)
        assert console.print.call_count == global_config.max_retries
        assert "Rate limit exceeded. Retrying after 60.0 seconds." in console.print.call_args.args[1]
        assert exc_info.value.message == "Too many requests"

    def test_create_function_401(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/functions"
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function")
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, url, status=401, json={"error": "Unauthorized"})
            with pytest.raises(CogniteProjectAccessError):
                client.functions.create_with_429_retry(fun)

            post_requests = [call for call in rsps.calls if call.request.url == url and call.request.method == "POST"]
            assert len(post_requests) == 1

    def test_create_function_invalid_json_float(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        fun = FunctionWrite(name="test_function", file_id=123, external_id="test_function", cpu=float("inf"))
        with pytest.raises(ValueError) as exc_info:
            client.functions.create_with_429_retry(fun)

        assert "Out of range float values are not JSON compliant" in str(exc_info.value)

    def test_create_function_invalid_json(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        fun = FunctionWrite(
            name="test_function",
            file_id=123,
            external_id="test_function",
            cpu=43j,  # Complex number, which is not JSON serializable
        )
        with pytest.raises(TypeError) as exc_info:
            client.functions.create_with_429_retry(fun)

        assert "Object 43j of type <class 'complex'> can't be serialized by the JSON encoder" in str(exc_info.value)



import pytest
from cognite.client.data_classes.raw import Row, RowList, RowWriteList

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.storageio import RawIO
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


@pytest.fixture()
def some_raw_tables() -> RowList:
    """Fixture to provide a sample RowList for testing."""
    return RowList(
        [
            Row(
                key=f"row{i}",
                columns={
                    "column1": f"value1_{i}",
                    "column2": f"value2_{i}",
                    "column3": f"value3_{i}",
                    "column4": {"nested_key": f"nested_value_{i}"},
                },
            )
            for i in range(100)
        ]
    )


class TestRawStorageIO:
    def test_download_upload(self, some_raw_tables: RowList) -> None:
        table = RawTable("test_db", "test_table")
        with monkeypatch_toolkit_client() as client:
            client.raw.rows.return_value = chunker(some_raw_tables, 10)
            io = RawIO(client)

            assert io.count(table) is None

            source = io.download_iterable(table, limit=100)
            json_chunks: list[list[dict[str, JsonVal]]] = []
            for chunk in source:
                json_chunk = io.data_to_json_chunk(chunk)
                assert isinstance(json_chunk, list)
                assert len(json_chunk) == 10
                for item in json_chunk:
                    assert isinstance(item, dict)
                json_chunks.append(json_chunk)
            data_chunks = (io.json_chunk_to_data(chunk) for chunk in json_chunks)
            for data_chunk in data_chunks:
                io.upload_items(data_chunk, table)

            assert client.raw.rows.insert.call_count == 10  # 100 rows in chunks of 10
            uploaded_rows = RowWriteList([])
            for call in client.raw.rows.insert.call_args_list:
                uploaded_rows.extend(call[1].get("row", []))

            assert uploaded_rows.dump() == some_raw_tables.as_write().dump()



from uuid import uuid4

from cognite.client.exceptions import CogniteException

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    CANVAS_INSTANCE_SPACE,
    Canvas,
    ContainerReferenceApply,
    FdmInstanceContainerReferenceApply,
)
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource
from cognite_toolkit._cdf_tk.exceptions import ToolkitMigrationError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.interactive_select import InteractiveCanvasSelect

from .base import BaseMigrateCommand
from .data_model import INSTANCE_SOURCE_VIEW_ID


class MigrationCanvasCommand(BaseMigrateCommand):
    canvas_schema_space = Canvas.get_source().space
    # Note sequences are not supported in Canvas, so we do not include them here.
    asset_centric_resource_types = frozenset({"asset", "event", "file", "timeseries"})

    def migrate_canvas(
        self,
        client: ToolkitClient,
        external_ids: list[str] | None = None,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> None:
        self.validate_access(
            client,
            instance_spaces=[CANVAS_INSTANCE_SPACE],
            schema_spaces=[self.canvas_schema_space, INSTANCE_SOURCE_VIEW_ID.space],
        )
        self.validate_migration_model_available(client)
        external_ids = external_ids or InteractiveCanvasSelect(client).select_external_ids()
        if external_ids is None or not external_ids:
            self.console("No canvases selected for migration.")
            return
        action = "Would migrate" if dry_run else "Migrating"
        self.console(f"{action} {len(external_ids)} canvases.")
        for external_id in external_ids:
            self._migrate_single_canvas(client, external_id, dry_run=dry_run, verbose=verbose)

    def _migrate_single_canvas(
        self,
        client: ToolkitClient,
        external_id: str,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> None:
        canvas = client.canvas.industrial.retrieve(external_id=external_id)
        if canvas is None:
            self.warn(MediumSeverityWarning(f"Canvas with external ID '{external_id}' not found. Skipping.. "))
            return
        update = canvas.as_write()
        to_migrate = [
            ref
            for ref in update.container_references
            if ref.container_reference_type in self.asset_centric_resource_types
        ]
        if not to_migrate:
            self.warn(
                LowSeverityWarning(
                    f"Canvas with name '{canvas.canvas.name}' does not have any asset-centric references. Skipping.. "
                )
            )
        if verbose:
            self.console(f"Found canvas: {canvas.canvas.name}")
        reference_ids = [ref.as_asset_centric_id() for ref in to_migrate]
        instance_sources = client.migration.instance_source.retrieve(reference_ids)
        source_by_reference_id = {source.as_asset_centric_id(): source for source in instance_sources}
        missing = set(reference_ids) - set(source_by_reference_id.keys())
        if missing:
            self.warn(
                HighSeverityWarning(
                    f"Canvas '{canvas.canvas.name}' has references to resources that are not been migrated: {humanize_collection(missing)}. Skipping.. "
                )
            )
            return
        if dry_run:
            self.console(
                f"Canvas '{canvas.canvas.name}' is ready for migration all {len(instance_sources)} references asset-centric resources found."
            )
            return
        if verbose:
            self.console(
                f"Migrating canvas '{canvas.canvas.name}' with {len(instance_sources)} references to asset-centric resources."
            )
        backup = canvas.as_write().create_backup()

        update.container_references = [
            ref
            for ref in update.container_references
            if ref.container_reference_type not in self.asset_centric_resource_types
        ]
        for ref in to_migrate:
            source = source_by_reference_id[ref.as_asset_centric_id()]
            fdm_ref = self.migrate_container_reference(ref, source, canvas.canvas.external_id)
            update.fdm_instance_container_references.append(fdm_ref)

        try:
            client.canvas.industrial.create(backup)
        except CogniteException as e:
            raise ToolkitMigrationError(f"Failed to create backup for canvas '{canvas.canvas.name}': {e!s}. ") from e
        try:
            client.canvas.industrial.update(update)
        except CogniteException as e:
            raise ToolkitMigrationError(
                f"Failed to migrate canvas '{canvas.canvas.name}': {e!s}. A backup was created with external ID '{backup.canvas.external_id}'."
            ) from e
        else:
            self.console(
                f'Canvas "{canvas.canvas.name}" migrated successfully with {len(to_migrate)} references to data model instances.'
            )

    @classmethod
    def migrate_container_reference(
        cls, reference: ContainerReferenceApply, source: InstanceSource, canvas_external_id: str
    ) -> FdmInstanceContainerReferenceApply:
        """Migrate a single container reference by replacing the asset-centric ID with the data model instance ID."""
        consumer_view = source.consumer_view()
        new_external_id = f"{canvas_external_id}_{uuid4()}"
        return FdmInstanceContainerReferenceApply(
            external_id=new_external_id,
            id_=reference.id_,
            container_reference_type="fdmInstance",
            instance_space=source.space,
            instance_external_id=source.external_id,
            view_space=consumer_view.space,
            view_external_id=consumer_view.external_id,
            view_version=consumer_view.version,
            label=reference.label,
            properties_=reference.properties_,
            x=reference.x,
            y=reference.y,
            width=reference.width,
            height=reference.height,
            max_width=reference.max_width,
            max_height=reference.max_height,
        )



import warnings
from collections.abc import Sequence
from itertools import groupby
from typing import overload

from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.data_modeling import (
    Node,
    NodeApplyResult,
    NodeApplyResultList,
    NodeId,
    NodeList,
    filters,
    query,
)
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.migration import (
    AssetCentricId,
    InstanceSource,
    ViewSource,
    ViewSourceApply,
)
from cognite_toolkit._cdf_tk.constants import COGNITE_MIGRATION_SPACE
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence

from .extended_data_modeling import ExtendedInstancesAPI


class InstanceSourceAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = InstanceSource.get_source()

    def retrieve(self, ids: Sequence[AssetCentricId]) -> NodeList[InstanceSource]:
        """Retrieve a list of instance sources by their IDs.

        Args:
            ids (Sequence[AssetCentricId]): A sequence of AssetCentricId objects representing the IDs of the instance sources to retrieve.
        """
        results: NodeList[InstanceSource] = NodeList[InstanceSource]([])
        for chunk in chunker_sequence(ids, self._RETRIEVE_LIMIT):
            retrieve_query = query.Query(
                with_={
                    "instanceSource": query.NodeResultSetExpression(
                        filter=filters.And(filters.HasData(views=[self._view_id]), self._create_dms_filter(ids)),
                        limit=len(chunk),
                    ),
                },
                select={"instanceSource": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            results.extend([InstanceSource._load(item.dump()) for item in chunk_response.get("instanceSource", [])])
        return results

    @staticmethod
    def _create_dms_filter(ids: Sequence[AssetCentricId]) -> filters.Filter:
        """Create a filter that matches all the AssetCentricIds in the list."""
        if not ids:
            raise ValueError("Cannot create a filter from an empty AssetCentricIdList.")
        to_or_filters: list[filters.Filter] = []
        instance_source_view = InstanceSource.get_source()
        for resource_type, resource_ids in groupby(
            sorted(ids, key=lambda x: x.resource_type), key=lambda x: x.resource_type
        ):
            is_resource = filters.Equals(instance_source_view.as_property_ref("resourceType"), resource_type)
            is_id = filters.In(
                instance_source_view.as_property_ref("id"), [resource_id.id_ for resource_id in resource_ids]
            )
            to_or_filters.append(filters.And(is_resource, is_id))
        return filters.Or(*to_or_filters)


class ViewSourceAPI:
    """API for retrieving view sources from the data model."""

    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self._instance_api = instance_api
        self._view_id = ViewSource.get_source()

    @overload
    def upsert(
        self,
        item: ViewSourceApply,
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> NodeApplyResult: ...

    @overload
    def upsert(
        self,
        item: Sequence[ViewSourceApply],
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> NodeApplyResultList: ...

    def upsert(
        self,
        item: ViewSourceApply | Sequence[ViewSourceApply],
        skip_on_version_conflict: bool = False,
        replace: bool = False,
    ) -> NodeApplyResult | NodeApplyResultList:
        """Upsert one or more view sources."""
        result = self._instance_api.apply(
            item, skip_on_version_conflict=skip_on_version_conflict, replace=replace
        ).nodes
        if isinstance(item, ViewSourceApply):
            return result[0]
        return NodeApplyResultList(result)

    @overload
    def retrieve(self, external_id: str) -> ViewSource | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> NodeList[ViewSource]: ...

    def retrieve(self, external_id: str | SequenceNotStr[str]) -> ViewSource | NodeList[ViewSource] | None:
        """Retrieve one or more view sources by their external IDs."""
        if isinstance(external_id, str):
            return self._instance_api.retrieve_nodes(NodeId(COGNITE_MIGRATION_SPACE, external_id), node_cls=ViewSource)
        else:
            nodes = self._instance_api.retrieve(
                nodes=[NodeId(COGNITE_MIGRATION_SPACE, ext_id) for ext_id in external_id], sources=[self._view_id]
            ).nodes
            return self._safe_convert(nodes)

    @overload
    def delete(self, external_id: str) -> NodeId: ...

    @overload
    def delete(self, external_id: SequenceNotStr[str]) -> list[NodeId]: ...

    def delete(self, external_id: str | SequenceNotStr[str]) -> NodeId | list[NodeId]:
        """Delete a view source or a list of view sources by their external IDs."""
        if isinstance(external_id, str):
            return self._instance_api.delete(NodeId(COGNITE_MIGRATION_SPACE, external_id)).nodes[0]
        else:
            return self._instance_api.delete([NodeId(COGNITE_MIGRATION_SPACE, ext_id) for ext_id in external_id]).nodes

    def list(self, resource_type: str | None = None, limit: int | None = DEFAULT_LIMIT_READ) -> NodeList[ViewSource]:
        """List view sources optionally filtered by resource type"""
        is_selected: filters.Filter | None = None
        if resource_type:
            is_selected = filters.Equals(self._view_id.as_property_ref("resourceType"), resource_type)

        nodes = self._instance_api.list(
            instance_type="node", filter=is_selected, limit=limit, space=COGNITE_MIGRATION_SPACE, sources=self._view_id
        )
        return self._safe_convert(nodes)

    @classmethod
    def _safe_convert(cls, nodes: NodeList[Node]) -> NodeList[ViewSource]:
        results = NodeList[ViewSource]([])
        for node in nodes:
            try:
                loaded = ViewSource._load(node.dump())
            except ValueError as e:
                warnings.warn(
                    HighSeverityWarning(
                        f"Node {node.as_id()!r} is in an invalid format. Skipping it. Error: {e!s}",
                    )
                )
                continue
            results.append(loaded)
        return results


class MigrationAPI:
    def __init__(self, instance_api: ExtendedInstancesAPI) -> None:
        self.instance_source = InstanceSourceAPI(instance_api)
        self.view_source = ViewSourceAPI(instance_api)



from collections.abc import Iterator
from pathlib import Path

import pytest
from cognite.client.data_classes import AssetList, AssetWrite, AssetWriteList
from cognite.client.data_classes.data_modeling import NodeId, Space
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteAsset
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import MigrateAssetsCommand
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.fixture()
def three_assets(toolkit_client: ToolkitClient, toolkit_space: Space) -> Iterator[AssetList]:
    client = toolkit_client
    space = toolkit_space.space
    assets = AssetWriteList([])
    for i in range(3):
        asset = AssetWrite(
            external_id=f"toolkit_asset_test_migration_{i}_{RUN_UNIQUE_ID}",
            name=f"toolkit_asset_test_migration_{i}_{RUN_UNIQUE_ID}",
            parent_external_id=f"toolkit_asset_test_migration_{0}_{RUN_UNIQUE_ID}" if i > 0 else None,
        )
        assets.append(asset)
    output = client.assets.retrieve_multiple(external_ids=assets.as_external_ids(), ignore_unknown_ids=True)
    if output:
        try:
            client.assets.delete(external_id=output.as_external_ids(), ignore_unknown_ids=True)
        except CogniteAPIError:
            client.data_modeling.instances.delete([NodeId(space, ts.external_id) for ts in output])
    created = client.assets.create(assets)

    yield created

    # Cleanup after test
    deleted = client.data_modeling.instances.delete([NodeId(space, ts.external_id) for ts in created])
    if deleted.nodes:
        return
    client.assets.delete(external_id=created.as_external_ids())


@pytest.mark.skip(
    "We will change the migration model and then we cannot have an integration tests deploying it while we are changing it."
)
class TestMigrateAssetsCommand:
    def test_migrate_assets_command(
        self,
        toolkit_client: ToolkitClient,
        three_assets: AssetList,
        toolkit_space: Space,
        tmp_path: Path,
    ) -> None:
        client = toolkit_client
        space = toolkit_space.space

        input_file = tmp_path / "timeseries_migration.csv"
        with input_file.open("w", encoding="utf-8") as f:
            f.write(
                "id,dataSetId,space,externalId\n"
                + "\n".join(
                    f"{a.id},{a.data_set_id if a.data_set_id else ''},{space},{a.external_id}" for a in three_assets
                )
                + "\n"
            )

        cmd = MigrateAssetsCommand(skip_tracking=True, silent=True)
        cmd.migrate_assets(
            client=client,
            mapping_file=input_file,
            dry_run=False,
            verbose=False,
            max_workers=2,
            output_dir=tmp_path,
        )
        node_ids = [NodeId(space, a.external_id) for a in three_assets]
        migrated_assets = client.data_modeling.instances.retrieve_nodes(node_ids, CogniteAsset)
        assert len(migrated_assets) == len(three_assets), "Not all assets were migrated successfully."

        csv_files = list(tmp_path.rglob("*MigrationResults.csv"))
        assert len(csv_files) == 1, "Expected one CSV file to be created."
        csv_file = csv_files[0]
        content = csv_file.read_text(encoding="utf-8")
        assert (
            content
            == f"""ResourceType,NodeSpace,NodeExternalId,ResponseStatus,ResponseMessage
asset,toolkit_test_space,toolkit_asset_test_migration_0_{RUN_UNIQUE_ID},200,
asset,toolkit_test_space,toolkit_asset_test_migration_1_{RUN_UNIQUE_ID},200,
asset,toolkit_test_space,toolkit_asset_test_migration_2_{RUN_UNIQUE_ID},200,
"""
        )



from collections.abc import Iterable
from pathlib import Path

from cognite.client.data_classes import RowList, RowWrite, RowWriteList
from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawDatabaseList, RawTable, RawTableList
from cognite_toolkit._cdf_tk.cruds import RawDatabaseCRUD, RawTableCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils.file import find_adjacent_files, read_yaml_file
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import StorageIO, StorageIOConfig


class RawIO(StorageIO[RawTable, RawTable, RowWriteList, RowList]):
    folder_name = "raw"
    kind = "RawRows"
    display_name = "Raw Rows"
    supported_download_formats = frozenset({".yaml", ".ndjson"})
    supported_compressions = frozenset({".gz"})
    supported_read_formats = frozenset({".parquet", ".csv", ".ndjson", ".yaml"})
    chunk_size = 10_000

    def as_id(self, item: dict[str, JsonVal] | object) -> RawTable:
        raise ValueError("You cannot extract an ID from a Raw Table row. Use a RawTable selector instead.")

    def count(self, selector: RawTable) -> int | None:
        # Raw tables do not support aggregation queries, so we do not know the count
        # up front.
        return None

    def download_iterable(self, selector: RawTable, limit: int | None = None) -> Iterable[RowList]:
        yield from self.client.raw.rows(
            db_name=selector.db_name,
            table_name=selector.table_name,
            limit=limit,
            # We cannot use partitions here as it is not thread safe. This spawn multiple threads
            # that are not shut down until all data is downloaded. We need to be able to abort.
            partitions=None,
            chunk_size=self.chunk_size,
        )

    def upload_items(self, data_chunk: RowWriteList, selector: RawTable) -> None:
        self.client.raw.rows.insert(db_name=selector.db_name, table_name=selector.table_name, row=data_chunk)

    def data_to_json_chunk(self, data_chunk: RowList) -> list[dict[str, JsonVal]]:
        return [row.as_write().dump() for row in data_chunk]

    def json_chunk_to_data(self, data_chunk: list[dict[str, JsonVal]]) -> RowWriteList:
        return RowWriteList([RowWrite._load(row) for row in data_chunk])

    def configurations(self, selector: RawTable) -> Iterable[StorageIOConfig]:
        yield StorageIOConfig(kind=RawTableCRUD.kind, folder_name=RawTableCRUD.folder_name, value=selector.dump())

    def load_selector(self, datafile: Path) -> RawTable:
        config_files = find_adjacent_files(datafile, suffix=f".{RawTableCRUD.kind}.yaml")
        if not config_files:
            raise ToolkitValueError(f"No configuration file found for {datafile.as_posix()!r}")
        if len(config_files) > 1:
            raise ToolkitValueError(f"Multiple configuration files found for {datafile.as_posix()!r}: {config_files}")
        config_file = config_files[0]
        loader = RawTableCRUD.create_loader(self.client)
        return loader.load_resource(read_yaml_file(config_file, expected_output="dict"))

    def ensure_configurations(self, selector: RawTable, console: Console | None = None) -> None:
        """Ensure that the Raw table exists in CDF."""
        db_loader = RawDatabaseCRUD.create_loader(self.client, console=console)
        db = RawDatabase(db_name=selector.db_name)
        if not db_loader.retrieve([db]):
            db_loader.create(RawDatabaseList([db]))
            if console:
                console.print(f"Created raw database: [bold]{db.db_name}[/bold]")

        table_loader = RawTableCRUD.create_loader(self.client, console=console)
        if not table_loader.retrieve([selector]):
            table_loader.create(RawTableList([selector]))
            if console:
                console.print(f"Created raw table: [bold]{selector.db_name}.{selector.table_name}[/bold]")



from collections.abc import Hashable
from datetime import date, datetime
from typing import Any, Literal, TypeAlias, TypeVar, get_args

from cognite.client.data_classes._base import WriteableCogniteResourceList

JsonVal: TypeAlias = None | str | int | float | bool | dict[str, "JsonVal"] | list["JsonVal"]

AssetCentricDestinationType: TypeAlias = Literal["assets", "files", "events", "timeseries", "sequences"]
AssetCentric: TypeAlias = Literal["asset", "file", "event", "timeseries", "sequence"]
DataType: TypeAlias = Literal["string", "integer", "float", "boolean", "json", "date", "timestamp"]
PythonTypes: TypeAlias = str | int | float | bool | datetime | date | dict[str, Any] | list[Any]

AVAILABLE_DATA_TYPES: set[DataType] = set(get_args(DataType))

T_ID = TypeVar("T_ID", bound=Hashable)
T_Selector = TypeVar("T_Selector", bound=Hashable)
T_WritableCogniteResourceList = TypeVar("T_WritableCogniteResourceList", bound=WriteableCogniteResourceList)
T_Value = TypeVar("T_Value")



from .assets import MigrateAssetsCommand
from .canvas import MigrationCanvasCommand
from .files import MigrateFilesCommand
from .prepare import MigrationPrepareCommand
from .timeseries import MigrateTimeseriesCommand

__all__ = [
    "MigrateAssetsCommand",
    "MigrateFilesCommand",
    "MigrateTimeseriesCommand",
    "MigrationCanvasCommand",
    "MigrationPrepareCommand",
]



from pathlib import Path

import questionary
from cognite.client.data_classes.capabilities import (
    Capability,
    DataSetScope,
    FilesAcl,
)
from cognite.client.data_classes.data_modeling import ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFileApply
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.panel import Panel
from rich.progress import track

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.extended_filemetdata import ExtendedFileMetadata
from cognite_toolkit._cdf_tk.exceptions import (
    AuthenticationError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence

from .base import BaseMigrateCommand
from .data_classes import MigrationMappingList


class MigrateFilesCommand(BaseMigrateCommand):
    cdf_cdm = "cdf_cdm"
    view_id = ViewId(cdf_cdm, "CogniteFile", "v1")
    chunk_size = 1000

    def source_acl(self, data_set_ids: list[int]) -> Capability:
        return FilesAcl(
            actions=[FilesAcl.Action.Read, FilesAcl.Action.Write],
            scope=DataSetScope(data_set_ids),
        )

    def migrate_files(
        self,
        client: ToolkitClient,
        mapping_file: Path,
        dry_run: bool = False,
        verbose: bool = False,
        auto_yes: bool = False,
    ) -> None:
        """Migrate resources from Asset-Centric to data modeling in CDF."""
        mappings = MigrationMappingList.read_mapping_file(mapping_file, "file")
        self.validate_access(
            client,
            instance_spaces=list(mappings.spaces()),
            schema_spaces=[self.cdf_cdm],
            data_set_ids=list(mappings.get_data_set_ids()),
        )
        self._validate_files(client, mappings)
        self.validate_available_capacity(client, len(mappings))

        if dry_run:
            self.console(f"Dry run mode. Would have migrated {len(mappings):,} Files to CogniteFiles.")
            return
        if not auto_yes and self._confirm(mappings) is False:
            return
        self._migrate(client, mappings, verbose)

    def _validate_files(self, client: ToolkitClient, mappings: MigrationMappingList) -> None:
        total_validated = 0
        chunk: MigrationMappingList
        for chunk in track(
            chunker_sequence(mappings, size=self.chunk_size),
            description="Validating...",
            total=len(mappings) // self.chunk_size + 1,
        ):
            try:
                files = client.files.retrieve_multiple(
                    ids=chunk.get_ids(),
                    ignore_unknown_ids=True,
                )
            except CogniteAPIError as e:
                raise AuthenticationError(
                    f"Failed to retrieve Files. This is likely due to lack of permissions: {e!s}"
                ) from e

            missing_count = len(files) - len(mappings)
            if missing_count > 0:
                raise ToolkitValueError(f"{missing_count} Files are missing in CDF.")

            missing_file_content = [file for file in files if file.uploaded is not True]
            if missing_file_content:
                raise ToolkitValueError(
                    f"The following files does not have file content yet: {humanize_collection(missing_file_content)}. "
                    "You can only migrate files that have file content uploaded."
                )

            existing_result = client.data_modeling.instances.retrieve(chunk.as_node_ids())
            if len(existing_result.nodes) != 0:
                raise ToolkitValueError(
                    "Some of the Files you are trying to migrate already exist in Data Modeling. "
                    f"Please remove the following files from the mapping file {humanize_collection(existing_result.nodes.as_ids())}"
                )
            total_validated += len(files)
        print(
            f"Validated {total_validated:,} Files for migration. "
            f"{len(mappings):,} mappings provided in the mapping file."
        )

    @staticmethod
    def _confirm(mappings: MigrationMappingList) -> bool:
        print(
            Panel(
                f"[red]WARNING:[/red] This operation [bold]cannot be undone[/bold]! "
                f"{len(mappings):,} Files will linked to the new CogniteFiles. "
                "This linking cannot be undone",
                style="bold",
                title="Migrate asset-centric Files to CogniteFiles",
                title_align="left",
                border_style="red",
                expand=False,
            )
        )

        if not questionary.confirm("Are you really sure you want to continue?", default=False).ask():
            print("Migration cancelled by user.")
            return False
        return True

    def _migrate(self, client: ToolkitClient, mappings: MigrationMappingList, verbose: bool) -> None:
        print("Migrating Files to CogniteFiles...")
        total_migrated = 0
        for chunk in track(
            chunker_sequence(mappings, size=self.chunk_size),
            description="Migrating Files to CogniteFiles...",
            total=len(mappings) // self.chunk_size + 1,
        ):
            if verbose:
                print(f"Migrating {len(chunk):,} Files...")

            # Set pending IDs for the chunk of mappings
            try:
                pending_files = client.files.set_pending_ids(chunk.as_pending_ids())
            except CogniteAPIError as e:
                raise ToolkitValueError(f"Failed to set pending IDs for Files: {e!s}") from e

            # The ExtendedTimeSeriesList is iterating ExtendedTimeSeries objects.
            converted_files = [self.as_cognite_file(file) for file in pending_files]  # type: ignore[arg-type]
            try:
                created = client.data_modeling.instances.apply_fast(converted_files)
            except CogniteAPIError as e:
                raise ToolkitValueError(f"Failed to apply Files: {e!s}") from e
            if verbose:
                print(f"Created {len(created):,} CogniteFiles.")
            total_migrated += len(created)
        print(f"Successfully migrated {total_migrated:,} Files to CogniteFiles.")

    @classmethod
    def as_cognite_file(cls, file: ExtendedFileMetadata) -> CogniteFileApply:
        if file.pending_instance_id is None:
            raise ToolkitValueError("ExtendedFiles must have a pending_instance_id set before migration.")

        return CogniteFileApply(
            space=file.pending_instance_id.space,
            external_id=file.pending_instance_id.external_id,
            name=file.name,
            mime_type=file.mime_type,
        )



from collections import Counter
from collections.abc import Iterable
from functools import partial
from pathlib import Path

from cognite.client.data_classes._base import T_CogniteResourceList
from rich.console import Console

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio import StorageIO, TableStorageIO
from cognite_toolkit._cdf_tk.utils.file import safe_write, to_directory_compatible, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.fileio import TABLE_WRITE_CLS_BY_FORMAT, Compression, FileWriter, SchemaColumn
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.useful_types import T_ID, JsonVal, T_Selector, T_WritableCogniteResourceList

from ._base import ToolkitCommand


class DownloadCommand(ToolkitCommand):
    def download(
        self,
        selectors: Iterable[T_Selector],
        io: StorageIO[T_ID, T_Selector, T_CogniteResourceList, T_WritableCogniteResourceList],
        output_dir: Path,
        verbose: bool,
        file_format: str,
        compression: str,
        limit: int | None = 100_000,
    ) -> None:
        """Downloads data from CDF to the specified output directory.

        Args:
            selectors: The selectors of the resources to download.
            io: The StorageIO instance that defines how to download and process the data.
            output_dir: The directory where the downloaded files will be saved.
            verbose: If True, prints detailed information about the download process.
            file_format: The format of the files to be written (e.g., ".ndjson").
            compression: The compression method to use for the downloaded files (e.g., "none", "gzip").
            limit: The maximum number of items to download for each selected set. If None, all items will be downloaded.
        """
        target_directory = output_dir / io.folder_name
        target_directory.mkdir(parents=True, exist_ok=True)
        compression_cls = Compression.from_name(compression)

        console = Console()
        filestem_counter: dict[str, int] = Counter()
        for selector in selectors:
            if verbose:
                console.print(f"Downloading {io.display_name} '{selector!s}' to {target_directory.as_posix()!r}")

            filestem = to_directory_compatible(str(selector))
            if filestem_counter[filestem] > 0:
                filestem = f"{filestem}_{filestem_counter[filestem]}"
            filestem_counter[filestem] += 1
            iteration_count = self._get_iteration_count(io, selector, limit)

            columns: list[SchemaColumn] | None = None
            if file_format in TABLE_WRITE_CLS_BY_FORMAT and isinstance(io, TableStorageIO):
                columns = io.get_schema(selector)
            elif file_format in TABLE_WRITE_CLS_BY_FORMAT:
                raise ToolkitValueError(
                    f"Cannot download {io.kind} in {file_format!r} format. The {io.kind!r} storage type does not support table schemas."
                )

            with FileWriter.create_from_format(
                file_format, target_directory, io.kind, compression_cls, columns=columns
            ) as writer:
                executor = ProducerWorkerExecutor[T_WritableCogniteResourceList, list[dict[str, JsonVal]]](
                    download_iterable=io.download_iterable(selector, limit),
                    process=io.data_to_json_chunk,
                    write=partial(writer.write_chunks, filestem=filestem),
                    iteration_count=iteration_count,
                    # Limit queue size to avoid filling up memory before the workers can write to disk.
                    max_queue_size=8 * 10,  # 8 workers, 10 items per worker
                    download_description=f"Downloading {selector!s}",
                    process_description="Processing",
                    write_description=f"Writing to {target_directory.as_posix()!r} in files with stem {filestem!r}",
                    console=console,
                )
                executor.run()
                executor.raise_on_error()
                file_count = writer.file_count

            for config in io.configurations(selector):
                config_file = output_dir / config.folder_name / f"{filestem}.{config.kind}.yaml"
                config_file.parent.mkdir(parents=True, exist_ok=True)
                safe_write(config_file, yaml_safe_dump(config.value))

            console.print(f"Downloaded {selector!s} to {file_count} file(s) in {target_directory.as_posix()!r}.")

    @staticmethod
    def _get_iteration_count(
        io: StorageIO[T_ID, T_Selector, T_CogniteResourceList, T_WritableCogniteResourceList],
        selector: T_Selector,
        limit: int | None,
    ) -> int | None:
        total = io.count(selector)
        if total is not None and limit is not None and total > limit:
            total = limit
        iteration_count: int | None = None
        if total is not None:
            iteration_count = total // io.chunk_size + (1 if total % io.chunk_size > 0 else 0)
        return iteration_count



from collections.abc import Iterable
from pathlib import Path

from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList, ChartWrite, ChartWriteList
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal, T_Selector

from ._base import StorageIO, StorageIOConfig
from ._selectors import AllChartSelector, ChartOwnerSelector, ChartSelector


class ChartIO(StorageIO[str, ChartSelector, ChartWriteList, ChartList]):
    folder_name = "cdf_application_data"
    kind = "Charts"
    display_name = "CDF Charts"
    supported_download_formats = frozenset({".ndjson"})
    supported_compressions = frozenset({".gz"})
    supported_read_formats = frozenset({".ndjson"})
    chunk_size = 10

    def as_id(self, item: dict[str, JsonVal] | object) -> str:
        if isinstance(item, dict) and isinstance(item.get("externalId"), str):
            # MyPy checked above.
            return item["externalId"]  # type: ignore[return-value]
        if isinstance(item, ChartWrite | Chart):
            return item.external_id
        raise TypeError(f"Cannot extract ID from item of type {type(item).__name__!r}")

    def download_iterable(self, selector: ChartSelector, limit: int | None = None) -> Iterable[ChartList]:
        selected_charts = self.client.charts.list(visibility="PUBLIC")
        if isinstance(selector, AllChartSelector):
            ...
        elif isinstance(selector, ChartOwnerSelector):
            selected_charts = ChartList([chart for chart in selected_charts if chart.owner_id == selector.owner_id])
        else:
            raise ToolkitNotImplementedError(f"Unsupported selector type {type(selector).__name__!r} for ChartIO")

        if limit is not None:
            selected_charts = ChartList(selected_charts[:limit])
        for chunk in chunker_sequence(selected_charts, self.chunk_size):
            ts_ids_to_lookup = {
                ts_ref.ts_id
                for chart in chunk
                for ts_ref in chart.data.time_series_collection or []
                if ts_ref.ts_external_id is None and ts_ref.ts_id is not None
            }

            if ts_ids_to_lookup:
                retrieved_ts = self.client.time_series.retrieve_multiple(
                    ids=list(ts_ids_to_lookup), ignore_unknown_ids=True
                )
                id_to_external_id = {ts.id: ts.external_id for ts in retrieved_ts}

                for chart in chunk:
                    for ts_ref in chart.data.time_series_collection or []:
                        if ts_ref.ts_id in id_to_external_id:
                            ts_ref.ts_external_id = id_to_external_id[ts_ref.ts_id]
            yield chunk

    def count(self, selector: ChartSelector) -> int | None:
        # There is no way to get the count of charts up front.
        return None

    def upload_items(self, data_chunk: ChartWriteList, selector: ChartSelector) -> None:
        # Todo validate all references exist in CDF before uploading.
        raise ToolkitNotImplementedError("Uploading charts is not implemented yet.")

    def data_to_json_chunk(self, data_chunk: ChartList) -> list[dict[str, JsonVal]]:
        return [chart.as_write().dump() for chart in data_chunk]

    def json_chunk_to_data(self, data_chunk: list[dict[str, JsonVal]]) -> ChartWriteList:
        return ChartWriteList._load(data_chunk)

    def configurations(self, selector: ChartSelector) -> Iterable[StorageIOConfig]:
        # Charts does not have any configurations for its data.
        return []

    def load_selector(self, datafile: Path) -> ChartSelector:
        raise ToolkitNotImplementedError("Loading charts is not implemented yet.")

    def ensure_configurations(self, selector: T_Selector, console: Console | None = None) -> None:
        # Charts do not have any configurations to ensure.
        return None



from pathlib import Path
from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.storageio import AssetIO, RawIO
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class UploadApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.upload_main)
        self.command("raw")(self.upload_raw_cmd)
        self.command("assets")(self.upload_assets_cmd)

    @staticmethod
    def upload_main(ctx: typer.Context) -> None:
        """Commands to upload data to CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf upload --help[/] for more information.")
        return None

    @staticmethod
    def upload_raw_cmd(
        ctx: typer.Context,
        input_dir: Annotated[
            Path,
            typer.Argument(
                help="The directory containing the RAW tables to upload. "
                "If not specified, the current working directory will be used.",
                exists=True,
                file_okay=False,
                dir_okay=True,
                resolve_path=True,
            ),
        ],
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the command will not actually upload the data, but will print what would be uploaded.",
            ),
        ] = False,
        ensure_tables: Annotated[
            bool,
            typer.Option(
                "--ensure-tables",
                "-e",
                help="If set, the command will ensure that the RAW database and table exist in CDF before uploading.",
            ),
        ] = True,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will upload data from the input directory to CDF."""
        cmd = UploadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(
            lambda: cmd.upload(
                io=RawIO(client=client),
                input_dir=input_dir,
                ensure_configurations=ensure_tables,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def upload_assets_cmd(
        ctx: typer.Context,
        input_dir: Annotated[
            Path,
            typer.Argument(
                help="The directory containing the asset files to upload.",
                exists=True,
                file_okay=False,
                dir_okay=True,
                resolve_path=True,
            ),
        ],
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the command will not actually upload the data, but will print what would be uploaded.",
            ),
        ] = False,
        ensure_dataset_and_labels: Annotated[
            bool,
            typer.Option(
                "--ensure-dataset-and-labels",
                help="If set, the command will look for datasets and labels in adjacent folders and create them if they do not exist.",
            ),
        ] = True,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will upload assets from input directory to CDF."""
        cmd = UploadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(
            lambda: cmd.upload(
                io=AssetIO(client=client),
                input_dir=input_dir,
                ensure_configurations=ensure_dataset_and_labels,
                dry_run=dry_run,
                verbose=verbose,
            )
        )
